package com.lordjoe.utilities;
import com.lordjoe.lib.xml.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.lang.reflect.*;

/** 
* Sample Use
*     public void buildInterfaces()
*    {
*        try {
*            COMEnumBuilder TheBuilder = new COMEnumBuilder();
*            TheBuilder.buildCOMEnums("com.onewebsystems.eagle.outlook","Enum","com/microsoft/outlook");
*       }
*        catch(java.lang.Exception ex) {
*            ex.printStackTrace();
*        }
*    }
* 
*/
public class COMCodeGenerator
{
    private File m_InputDirectory;
    private File m_TargetDirectory;
    private String m_Package;
    private String m_InterfaceStart;
    private String m_Prefix;
    private String m_InterfacePrefix;
    private COMEnumBuilder m_EnumBuilder;
    private COMSurrogateBuilder m_SurrogateBuilder;
    
    public COMCodeGenerator(String Package,String TargetDirectory,String SourceDirectory,String Prefix,String InterfacePrefix)
    {
        m_InputDirectory = new File(SourceDirectory);
        if(!m_InputDirectory.exists())
            throw new IllegalArgumentException("File " + SourceDirectory + " does not exits");
        if(!m_InputDirectory.isDirectory())
            throw new IllegalArgumentException("File " + SourceDirectory + " is not a directory");


        m_TargetDirectory = new File(TargetDirectory);
        if(!m_TargetDirectory.exists())
            throw new IllegalArgumentException("File " + TargetDirectory + " does not exits");
        if(!m_TargetDirectory.isDirectory())
            throw new IllegalArgumentException("File " + TargetDirectory + " is not a directory");
            
        m_InterfaceStart = m_InputDirectory.getPath().replace('\\','/') + "/Ol";
        m_Package = Package;
        m_Prefix = Prefix;
        m_InterfacePrefix = InterfacePrefix;
        m_EnumBuilder = new COMEnumBuilder();
        m_SurrogateBuilder = new COMSurrogateBuilder();
    }
    
    
    public void buildCOMWrappers() throws IOException
    {
        String[] items = FileUtilities.getAllFilesWithExtension(m_InputDirectory,"java");
        String TestString = m_InputDirectory.getPath().replace('\\','/') + "/_";
        TestString = TestString.replace('\\','/');
        for(int i = 0; i < items.length; i++) {
            String ClassName = items[i].substring(0,items[i].length() - 5);
            ClassName = ClassName.replace('\\','/');
            if(!ClassName.startsWith(TestString)) {
                ClassName = ClassName.replace('/','.');
                Class test = ClassUtilities.classForName(ClassName);
                if(test != null) {
                    handleClass(test);
                }
            }
            else {
                ClassName = ClassName; // drop classes starting with _
            }
        }
    }
    
    protected void handleClass(Class TheClass) throws IOException
    {
        if(TheClass.isInterface())
            handleInterface(TheClass);
        else
            handleCOMClass(TheClass);
    }
    
    protected void handleInterface(Class TheClass) throws IOException
    {
        m_EnumBuilder.buildCOMEnum(m_Package,m_InterfacePrefix, "Enum", TheClass,m_TargetDirectory);
    }
    
    
    protected void handleCOMClass(Class TheClass) throws IOException
    {
        m_SurrogateBuilder.buildCOMSurrogate(m_Package, m_Prefix, TheClass,m_TargetDirectory);
    }
}