/**{ file
    @name OADBObject.java
    @function Base class for table based Objects
    @author Steven M. Lewis
    @version 1.0
    @date Mon Mar 10 13:28:03  1997
    @copyright
	************************
	*  Copyright (c) 1999
	*  Steven M. Lewis
	************************

}*/
package com.lordjoe.utilities;
import java.sql.*;
import java.util.*;
import java.io.*;


public abstract class JDBCUtilities
{
    public static final String DEFAULT_DB_PREFIX = "jdbc:cloudscape:";
    public static final String DEFAULT_DRIVER_NAME = "COM.cloudscape.core.JDBCDriver";
    
    protected static String gDBPrefix = DEFAULT_DB_PREFIX;
    protected static String gDriverName = DEFAULT_DRIVER_NAME;
    
    public static int SELECT_PRIVILEGE = 1;
    public static int INSERT_PRIVILEGE = 2;
    
    
    public static void setDBPrefix(String s)
    {
        gDBPrefix = s;
    }
    
    public static void setDriverName(String s)
    {
        gDriverName = s;
    }
    
    
	public static Connection getConnection(String DBName) throws SQLException
	{
	    try {
		    Class.forName(gDriverName);
		    String ConnectionName = gDBPrefix + DBName;
		 //   System.err.println("Trying to open Connection " + ConnectionName);
		    Connection connection =
			    DriverManager.getConnection(ConnectionName);
		    return(connection);
		}
		catch(ClassNotFoundException ex) {
		    throw new IllegalStateException(ex.toString());
		}
		catch(SQLException ex) {
		    throw ex; // debugging stop
		}
	}
	
    /** get a database Connection object 
    * @param  driver  non-null string JDBCDriver name
    * @Param  url     non-null string database url
    * @param  user    non-null string datbase user name
    * @param  passwd  non-null string user passwd
    * @return con     Database Connection object
    */
    public static Connection getConnection(String driver, String url, 
					   String user, String passwd)
	throws SQLException {
        try { 
            Class.forName(driver);
            Connection con = DriverManager.getConnection(url, user, passwd);
            return con;
        }
		catch(ClassNotFoundException ex) {
		    throw new IllegalStateException(ex.toString());
		
		}catch(SQLException ex) {
		    throw ex; // debugging stop
		}
    }              
	
	public static boolean hasTable(String TableName,Connection conn) throws SQLException
	{
	    DatabaseMetaData md = conn.getMetaData();
	    String[] Types =  { "TABLE", "VIEW" };
        ResultSet rs = md.getTables(null, null, TableName,null);
        boolean ret = rs.next();
        rs.close();
        if(ret) 
            return(ret);
            
        rs = md.getTables(null, null, TableName.toUpperCase(),null);
        ret = rs.next();
        rs.close();
        return(ret);
	}
	    
	
	public static String[] listTables(Connection conn) throws SQLException
	{
	    String[] Types =  { "TABLE", "VIEW" };
	    return(listTables(conn,Types));
	}
	    
	
	
	public static String[] listTables(Connection conn, String[] tableTypes) throws SQLException
	{
	    try {
	        DatabaseMetaData md = conn.getMetaData();
            ResultSet rs = md.getTables(null, null, "%",tableTypes);
            List holder = new ArrayList();
            while(rs.next()) {
                String Table = rs.getString(3); // gte the table
                holder.add(Table);
            }
	        String[] Tables = (String[])Util.collectionToArray(holder,String.class);
	        // return(accessableTables(conn,Tables));
	        return(Tables);
	    }
	    catch(java.lang.UnsatisfiedLinkError ex) {
	        String path = System.getProperty("java.library.path");
	        String Message = "Cannot library - path is\n" + path;
	        Message += "\nError is:" + ex.getMessage();
	        throw new IllegalStateException(Message);
	    }
	}
	
	protected static String[] accessableTables(Connection conn, String[] TableNames) throws SQLException
	{
        List holder = new ArrayList();
        for(int i = 0; i < TableNames.length; i++)
        {
            if(tableIsAccesable(conn,TableNames[i]))
                holder.add(TableNames[i]);
        }
       return((String[])Util.collectionToArray(holder,String.class));
	}
	
	protected static boolean tableIsAccesable(Connection conn, String TableName) throws SQLException
	{
        int NeedsPrivilege = SELECT_PRIVILEGE | INSERT_PRIVILEGE;
 	    DatabaseMetaData md = conn.getMetaData();
 	    try {
	        String me = md.getUserName();
            ResultSet rs = md.getTablePrivileges(null, null,TableName);
            boolean HasResults = false;
            while(rs.next()) {
                HasResults = true;
                String Grantee = rs.getString(5); // gte the table
                String Privilege = rs.getString(6); // gte the table
                if(Privilege.equalsIgnoreCase("public"))
                    NeedsPrivilege = 0;
                if(Privilege.equalsIgnoreCase("insert"))
                    NeedsPrivilege &= ~INSERT_PRIVILEGE;
                if(Privilege.equalsIgnoreCase("select"))
                    NeedsPrivilege &= ~SELECT_PRIVILEGE;
            }
            if(NeedsPrivilege == 0)
                return(true);
            return(!HasResults); // so if no results then is OK I guess
        }
        catch(SQLException ex) {
            return(true); // no way to tell
        }
	}
	
	
	    
	public static int generateKey(Statement statement,String TableName,String KeyColumn)  throws SQLException
	{
			String sql = "SELECT MAX(" + KeyColumn + ") FROM  " + TableName; 
			
			ResultSet rs = statement.executeQuery(sql);
			int ret = 0;
			if(!rs.next())
			    return(1);
			return(rs.getInt(1) + 1);
	}
    
    public static int keyExists(Statement statement,String TableName,String KeyColumn,int theKey)  throws SQLException
	{
			String sql = "select (max("+KeyColumn+")+1) from "+TableName+" where EXISTS(select "+KeyColumn+" from "+TableName+" where "+KeyColumn+" = "+theKey+")";
			ResultSet rs = statement.executeQuery(sql);
			int ret = 0;
			if(!rs.next())
			    return(theKey);
			ret = rs.getInt(1);
			if (ret <= 0 )
			    return theKey;
			return ret;
	}
	
	/**
	* Columns are designated as Catalog.Schema.Table.Column
	* This returns the Column part of the string
	* @param FullName non-null String as above
	* @return non-null column name
	*/
	public static String getColumnNamePart(String FullName)
	{
	    int index = FullName.lastIndexOf('.');
	    if(index == -1)
	        return(FullName);
	    return(FullName.substring(index + 1));
	}
}
