/**{ file
    @name LException.java
    @function root for all LordJoe exceptions - these derive from
         runtimeException and do not have to be declared
    @author> Steven M. Lewis
    @copyright> 
	************************
	*  Copyright (c) 1996,97,98
	*  Steven M. Lewis
	*  www.LordJoe.com
	************************

*  Updated to make it compile on Java 1.5
*  Miika Nurminen
*  
*    @date> Fri 5. May 2006
*    @version> 1.1
}*/ 
package com.lordjoe.utilities;
import java.util.*;
import java.io.*;
/**{ class
    @name LException
    @function root for all LordJoe exceptions - these derive from
         runtimeException and do not have to be declared
}*/ 
public class LException extends RuntimeException implements IStackTraceable
{
   private Throwable m_NestedException; // held exception
   private String    m_StackTrace; // marks where the exception was thrown 
   private String    m_DetailMessage;  // used instead of unsetable Throwable Message

    //- ******************* 
    //- Methods 
    /**{ constructor
        @name LException
        @function Constructor of LException
    }*/ 
    public LException() {
    }

    /**{ constructor
        @name LException
        @function Constructor of LException
        @param message String to attach
    }*/ 
    public LException(String message) {
        super(message);
    }

    /**{ constructor
        @name LException
        @function Constructor of LException
        @param message String to attach
    }*/ 
    public LException(Exception ex) {
        super(ex.getMessage());
        setNestedException(ex);
    }
    
    /**{ constructor
        @name LException
        @function Constructor of LException
        @param message String to attach
    }*/ 
    public LException(String s,Exception ex) {
        super(s);
        setNestedException(ex);
    }
   
   public Throwable getNestedException() {
     if(m_NestedException == null)
        return(this);
     else 
        return(m_NestedException);
   }
    
   public void setNestedException(Throwable ex) {
      if(ex == this)
        throw new IllegalArgumentException("Cannot nest exception into self");
      m_NestedException = ex;
      String nestedStackTrace = ""; //StackTrace.getStackTrace(ex);
      setStackTrace(StackTrace.mergeStackTrace(m_StackTrace,nestedStackTrace));
   }

/*   
   public String getStackTrace()
   { 
        return(m_StackTrace);
   }
*/
   public void setStackTrace(String s)
   { 
        m_StackTrace = s;
   }
   
   public String getDetailMessage()
   { 
        return(m_DetailMessage);
   }

   public void setDetailMessage(String s)
   { 
        m_DetailMessage = s;
   }
   

    /**
     * Returns the errort message string of this throwable object.
     *
     * @return  the error message string of this <code>Throwable</code> 
     *          object if it was {@link #Throwable(String) created} with an 
     *          error message string; or <code>null</code> if it was 
     *          {@link #Throwable() created} with no error message. 
     *            
     */
    public String getMessage() {
	String msg = super.getMessage();
        boolean hasSuperPassage = (msg == null);	
        StringBuffer sb = new StringBuffer();
        if(!hasSuperPassage)
            sb.append(msg);
	    Throwable nested = getNestedException();
        if (nested != null && nested != this) {
                sb.append("\n     Nested Exception: ");
	        sb.append(ClassUtilities.shortClassName(nested));
                if(!hasSuperPassage)
		    sb.append(" - ").append(nested.getMessage());
                sb.append('\n');
	    }
        if (getDetailMessage() != null) {
                sb.append(getDetailMessage()).append('\n');
	    }
        return(sb.toString());
    }

//- ******************* 
//- End Class LException
}
