package com.lordjoe.utilities;
import com.lordjoe.lib.xml.*;
import java.util.*;
import java.text.*;
import java.lang.reflect.*;

/**
* this builds an iterator frim an auromation object 
* which implements GetFirst,GetNext and thest return an Object
* which may be used to construct a target object
* @author S Lewis
*  Updated to make it compile on Java 1.5
*  Miika Nurminen
*  
*    @date> Fri 5. May 2006
*    @version> 1.1
*/
public class ReflectiveIterator implements Iterator
{
    private Object m_Target;
    private int m_count;
    private int m_index;
    private Method m_item; // Method called Item(int)
    private Method m_next;
    private Method m_first;
    Constructor[] m_Constructors;
    private Class m_ReturnClass;
    
    /**
    * @param target non-null auromation object implementing GetFirst,GetNext and GetCount
    * @param ReturnClass non-null class of object to be build from the return 
    * of get_Count,GetFirst and GetNext 
    */
    public ReflectiveIterator(Object target,Class ReturnClass) {
        setReturnClass(ReturnClass);
        setTarget(target);
    }
    
    /**
    * @param target non-null auromation object implementing GetFirst,GetNext and GetCount
    * @param ReturnClass non-null class of object to be build from the return 
    * of get_Count,GetFirst and GetNext 
    * @param start non-negative start position
    * @param end non-negative end position - note 0 means ignore start, show all
    */
    public ReflectiveIterator(Object target,Class ReturnClass,int start,int end) {
        this(target,ReturnClass);
        if(end == 0)
            return; // end == 0 says get all
        m_index = Math.max(0,start - 1);
        m_count = Math.min(end,m_count);
    }
    
    protected void setTarget(Object target) {
        m_Target = target;
        Class TargetClass = m_Target.getClass();
        try {
            Method CountMethod = TargetClass.getMethod("getCount",(Class[])null);
            Integer iCount = (Integer)invokeMethod(CountMethod);
            m_count = iCount.intValue();
            if(m_count == Integer.MAX_VALUE) {
               m_count = 0; // probably a good call
            }
            try {
                Class[] itemArgs = { Integer.TYPE };
                m_item = TargetClass.getMethod("getItem",itemArgs);
                return;
            }
            catch(NoSuchMethodException ex) {
                // OK try first and next
            }
            m_first = TargetClass.getMethod("GetFirst",(Class[])null);
            m_next = TargetClass.getMethod("GetNext",(Class[])null);
        }
        catch(NoSuchMethodException ex) {
            throw new IllegalArgumentException("Required methods - GetFirst,GetNext and GetCount do not exist in class " + TargetClass.getName());
        }
   }

    protected void setReturnClass(Class ReturnClass) {
       m_ReturnClass = ReturnClass;
        Class[] args = { Object.class };
        Constructor[] testConstructors = ReturnClass.getConstructors();
        List holder = new ArrayList();
        for(int i = 0; i < testConstructors.length; i++) {
            Class[] clsargs = testConstructors[i].getParameterTypes();
            if(clsargs.length == 1 && !clsargs[0].isPrimitive())
                holder.add(testConstructors[i]);
        }
        m_Constructors = new Constructor[holder.size()];
        holder.toArray(m_Constructors);
        if(m_Constructors.length == 0)
            throw new IllegalArgumentException("Required constructor(OBject) - does not exist in class " + m_ReturnClass.getName());
   }

	public boolean hasNext() {
	    return(m_index < m_count);
	}

	public Object next() {
	    if(!hasNext()) 
	        return(null);
	    Object item = null;
	    if(m_item != null) {
	        item = invokeMethod(m_item,m_index + 1); // item is 1 based
	    }
	    else {
	        if(m_index == 0)
	            item = invokeMethod(m_first);
	        else
	            item = invokeMethod(m_next);
	    }
	    m_index++;
	    return(buildTarget(item));
	}

	public void remove() {
	    throw new UnsupportedOperationException("Remove is not supported");
	}
	
	protected Object buildTarget(Object item) {
	    if(m_ReturnClass.isInstance(item))
	        return(item);
	    Object[] args = { item };
	    Constructor test = null;
	    try {
	       for(int i = 0; i < m_Constructors.length; i++) {
	         test = m_Constructors[i];
	         Class[] clsargs = test. getParameterTypes();
	         if(clsargs[0].isInstance(item)) {
	            Object ret = test.newInstance(args);
	            return(ret);
	         }
	       }
	        throw new IllegalStateException("No valid Constructor");
	    }
	    catch(IllegalAccessException ex ) {
	        ex.printStackTrace();
	        throw new IllegalStateException("Cannot access method " + test.getName());
	    }
	    catch(InvocationTargetException ex ) {
	        ex.printStackTrace();
	        throw new IllegalStateException("Cannot create Object");
	    }
	    catch(InstantiationException ex ) {
	        ex.printStackTrace();
	        throw new IllegalStateException("Cannot create Object");
	    }
	}
	
	protected Object invokeMethod(Method m) {
	    try {
	       Object ret = m.invoke(m_Target,(Object[])null);
	       return(ret);
	    }
	    catch(IllegalAccessException ex ) {
	        throw new IllegalStateException("Cannot access method " + m.getName());
	    }
	    catch(InvocationTargetException ex ) {
	        throw new IllegalStateException("Method had error");
	    }
	}
	protected Object invokeMethod(Method m,int index) {
	    try {
	       Object[] args = { new Integer(index) };
	       Object ret = m.invoke(m_Target,args);
	       return(ret);
	    }
	    catch(IllegalAccessException ex ) {
	        throw new IllegalStateException("Cannot access method " + m.getName());
	    }
	    catch(InvocationTargetException ex ) {
	        throw new IllegalStateException("Method had error");
	    }
	}
}
