package com.lordjoe.utilities;
import java.util.*;
import java.io.*;
//import javax.naming.*;

/**
* represents one item on the stack
*  Updated to make it compile on Java 1.5
*  Miika Nurminen
*  
*    @date> Fri 5. May 2006
*    @version> 1.1
*/
public class StackTrace 
{
    private Class        m_Exception;
    private StackElement[] m_Elements;
    
    public StackTrace()
    {
        this(getStackTrace());
    }
    
    public StackTrace(Throwable ex)
    {
/*        this(getStackTrace(ex));*/
        m_Exception = ex.getClass();
    }
    
    public StackTrace(String[] items)
    {
        List holder = new ArrayList();
        int i = 0;
        if(items.length > 0 && items[i].indexOf("(") == -1) 
            m_Exception = buildException(items[i++]);
            
        for(; i < items.length; i++) {
            holder.add(new StackElement(items[i]));
        }
        m_Elements = new StackElement[holder.size()];
        holder.toArray(m_Elements);
        
    }

    protected StackTrace(Class TheClass,StackElement[] items)
    {
        m_Exception = TheClass;
        m_Elements = items;
        
    }

    public StackTrace merge(StackTrace source)
    {
        List Holder = new ArrayList();
        StackElement[] SourceItems = source.m_Elements;
        int index = 0;
        for(int i = 0; i < SourceItems.length; i++) 
            Holder.add(SourceItems[i]);
            
        StackElement[] MergedItems = new StackElement[Holder.size()];
        Holder.toArray(MergedItems);
        return(new StackTrace(m_Exception,MergedItems));
    }
    
    /**
    * custom string 
    * @return non-null string similar to stack dump
    */
    public String toString()
    {
        StringBuffer st = new StringBuffer();
        if(m_Exception != null) {
            st.append(m_Exception.getName());
            st.append("\n");
        }
        for(int i = 0; i < m_Elements.length; i++) {
            st.append("at ");
            st.append(m_Elements[i].toString());
            st.append("\n");
        }
        return(st.toString());
       
    }
    
    public StackTrace(String TheStack)
    {
        this(TheStack != null ? Util.parseString(TheStack,"\n"): new String[0]);
    }
     /**
    * return the current 
    trace
    * @return exception - will be null if we generated the trace
    */
    public static Class buildException(String s)
    {
        try {
            Class ret = Class.forName(s);
            if(ret == TraceException.class)
                return(null); // ignore artificial exceptions
            return(ret);
        }
        catch(ClassNotFoundException ex) {
            return(null);
        }
    }
   
    /**
    * return the current 
    trace
    * @return non-null string representing the stack trace
    */
    public static String getStackTrace()
    {
      return "ERROR";
//        return(getStackTrace(new TraceException()));
    }
    
    public static String mergeStackTrace(String Enclosing,String source)
    {
        StackTrace EnclosingTrace = new StackTrace(Enclosing);
        StackTrace SourceTrace = new StackTrace(source);
        StackTrace merge = EnclosingTrace.merge(SourceTrace);
        return(merge.toString());
    }
        
    
    /**
    * special class so we cal read a stack trace
    */
    private static class TraceException extends Throwable {
    }
    
    /**
    * return the stack trace in the throwable
    * @param non-null Throwable holding the stack
    * @return non-null string representing the stack trace
    */
/*    public static String getStackTrace(Throwable ex)
    {    
        if(ex instanceof IStackTraceable) {
	   
            String tracestr = (((IStackTraceable)ex).getStackTrace());
	    if (tracestr != null ) {  // might be null -- MWR
		return tracestr;
	    }	   
	}

        String[]items = getStackTraceItems(ex);
        StringBuffer out = new StringBuffer();
        for(int i = 0; i < items.length - 1; i++) 
        {
            out.append(items[i]);
            if(i < items.length - 1) 
                out.append("\n");
        }
        return(out.toString());
    }*/
    /**
    * return the current stack trace
    * @retur non-null string representing the stack trace
    */
    public static String[] getStackTraceItems()
    {
        return(getStackTraceItems(new TraceException()));
    }
    
    /**
    * drop at at the beginning and any nonprinting chars at the end
    * @param in non-null string input
    * @return non-null output
    */
    protected static String filterStackTrace(String in) 
    {
        StringBuffer t = new StringBuffer(in);
        for(int i = t.length() - 1; i >= 0; i--) {
            if(Character.isLetterOrDigit(t.charAt(i)))
                return(t.toString());
        }
        return(t.toString().trim());
    }
     /**
    * drop at at the beginning and any nonprinting chars at the end
    * @param in non-null string input
    * @return non-null output
    */
    protected static String filterExceptionClass(String in) 
    {
        int ColonIndex = in.indexOf(":");
        if(ColonIndex == -1)
            return(in.trim());
        return(in.substring(0,ColonIndex).trim());
    }
    
   
    /**
    * return the  stack trace of the throwable
    * @return non-null string representing the stack trace
    */
    public static String[] getStackTraceItems(Throwable ex)
    {
        ByteArrayOutputStream bo = new ByteArrayOutputStream(1000);
        PrintStream p = new PrintStream(bo);
        ex.printStackTrace(p);
        String s = bo.toString();
        String[] items = Util.parseString(s,"\n");
        List holder = new ArrayList(items.length);
        int i = 0;
        holder.add(filterExceptionClass(items[i++]));
        int start = items.length - 1; 
        // find stack start
        for(; start >= 0; start--) {
            if(!StackElement.isStackElement(items[start])) {
                start++;
                break;
            }
        }
        // ignore the exceptions we generated
        if(ex instanceof TraceException)    
            start++;
            
        for(i = start; i <  items.length; i++) {
            if(StackElement.isStackElement(items[i]))
                holder.add(filterStackTrace(items[i].substring(3)));
        }
        
        String[] ret = new String[holder.size()];
        holder.toArray(ret);
        return(ret);
    }
    
    /**
    * code to get parameter Exception
    * @return <Add Comment Here>
    * @see getException
    */
    public Class getException()
    {
        return(m_Exception);
    }

    /**
    * code to set parameter Exception
    * @param in <Add Comment Here>
    * @see getException
    */
    public void setException(Class in)
    {
        m_Exception = in;
    }

    /**
    * code to get parameter Elements
    * @return <Add Comment Here>
    * @see getElements
    */
    public StackElement[] getElements()
    {
        return(m_Elements);
    }

    /**
    * code to set parameter Elements
    * @param in <Add Comment Here>
    * @see getElements
    */
    public void setElements(StackElement[] in)
    {
        m_Elements = in;
    }

}
