package com.lordjoe.utilities;
import java.util.*;

/**
* Collection of utilities for dealing with threads
* @author Steve Lewis
*/
public class ThreadUtilities
{
    public static final int WAIT_INCR = 10 * Util.ONE_SECOND;
    
    /**
    * waits forever
    * @throws runtime ThreadInterruptedException on interruption
    */
    public static void waitForever()
    {
        while(true) {
            waitFor(WAIT_INCR);
        }
    }
    
    /**
    * waits for millis milliseconds then returns
    * @param millis positive milliseconds to wait
    * @throws runtime ThreadInterruptedException on interruption
    */
    public static void waitFor(int millis)
    {
        if(millis < WAIT_INCR) {
            doWaitFor(millis);
        }
        else {
            doMultiWaitFor(millis);
        }
    }
    
    /**
    * waits for millis milliseconds then returns
    * @param millis positive milliseconds to wait
    * @throws runtime ThreadInterruptedException on interruption
    */
    protected static void doWaitFor(int millis)
    {
        try {
            Thread.sleep(millis);
        }
        catch(InterruptedException ex) {
            throw new ThreadInterruptedException("Watch unexpectedly interrupted");
        }
    }
    
     /**
    * waits for millis milliseconds then returns
    * @param millis positive milliseconds to wait
    * @throws runtime ThreadInterruptedException on interruption
    */
    protected static void doMultiWaitFor(int millis)
    {
        long Stop = SysTime.getTic() + millis;
        while(SysTime.getTic() < Stop) {
            doWaitFor(WAIT_INCR);
        }
    }
   
    /**
    * waits for millis milliseconds then returns
    * in this case the current thread will release locks on lock and
    * wait and will be restarted by notification
    * @param millis positive milliseconds to wait
    * @throws runtime ThreadInterruptedException on interruption
    */
    public static void waitFor(int millis,Object lock)
    {
        try {
            lock.wait(millis);
        }
        catch(InterruptedException ex) {
            throw new ThreadInterruptedException("Watch unexpectedly interrupted");
        }
    }

    /**
    * waits for a random interval which is at least MinWait
    * evenly distributes between MinWait and MinWait + AverageWait
    * @param MinWait positive milliseconds minimum wait
    * @param AverageWait positive milliseconds minimum wait
    * @throws runtime ThreadInterruptedException on interruption
    */
    public static void randomSleep(int MinWait,int AverageWait)
    {
        ThreadUtilities.waitFor(MinWait + Util.randomInt(AverageWait));
    }
}
