package com.lordjoe.utilities;
import com.lordjoe.lib.xml.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.lang.reflect.*;
/**
*  Updated to make it compile on Java 1.5
*  Miika Nurminen
*  
*    @date> Fri 5. May 2006
*    @version> 1.1
  */
public class TypelibDispatchBuilder
{
    public static final int NUM_NAMED_OPTIONAL_PARAMS = 2; // above this number optional TheParams add code 
                                                           // to call by name
                                                           
    private static Map gTypeToVariantOp = buildTypeToVariantOp();
    private static Map gKindToDispatchConstant = buildKindToDispatchConstant();
    private String m_Package;
    private String m_AddedImports;
    private File m_TargetDirectory;
    private Set m_DispatchClasses;
    private boolean m_doubleIsDate; // 
    private TypeLib m_TypeLib; // 
    
    protected static Map  buildTypeToVariantOp()
    {
        Map ret = new HashMap();
        ret.put("int","toInt");
        ret.put("boolean","toBoolean");
        ret.put("double","toDouble");
        ret.put("short","toShort");
        ret.put("float","toFloat");
        ret.put("String","getString");
        ret.put("byte","toByte");
        ret.put("Dispatch","toDispatch");
        return(ret);
    }

    
    protected static Map  buildKindToDispatchConstant()
    {
        Map ret = new HashMap();
        ret.put("IDispatch.PROPERTYGET","DispatchConstants.DISPATCH_PROPERTYGET");
        ret.put("IDispatch.PROPERTYPUT","DispatchConstants.DISPATCH_PROPERTYPUT");
        ret.put("IDispatch.INVOKE_PROPERTYPUTREF","DispatchConstants.DISPATCH_PROPERTYPUTREF");
        ret.put("INVOKE_PROPERTYPUTREF","DispatchConstants.DISPATCH_PROPERTYPUTREF");
        ret.put("IDispatch.METHOD","DispatchConstants.DISPATCH_METHOD");
        return(ret);
    }
    
    
    public TypelibDispatchBuilder()
    {
        m_doubleIsDate = true;
        m_AddedImports = "";
        m_Package = "";
    }
    
    public TypelibDispatchBuilder(TypeLib TheLib,String APackage,String AddedImports)
    {
        this();
        setTypeLib(TheLib);
        m_AddedImports = AddedImports;
        m_Package = APackage;
    }
    
    public void setDoubleIsDate(boolean doit) 
    {
        m_doubleIsDate = doit;
    }
    
    
    /**
    * code to get parameter TargetDirectory
    * @return <Add Comment Here>
    * @see getTargetDirectory
    */
    public File getTargetDirectory()
    {
        return(m_TargetDirectory);
    }

    /**
    * code to set parameter TypeLib
    * @param in <Add Comment Here>
    * @see getTypeLib
    */
    public void setTypeLib(TypeLib in)
    {
        m_TypeLib = in;
    }

    /**
    * code to get parameter TypeLib
    * @return <Add Comment Here>
    * @see getTypeLib
    */
    public TypeLib getTypeLib()
    {
        return(m_TypeLib);
    }

    /**
    * code to set parameter TargetDirectory
    * @param in <Add Comment Here>
    * @see getTargetDirectory
    */
    public void setTargetDirectory(File in)
    {
        m_TargetDirectory = in;
    }

   
     /**
    * code to get parameter Package
    * @return <Add Comment Here>
    * @see getPackage
    */
    public String getPackage()
    {
        return(m_Package);
    }

    /**
    * code to set parameter Package
    * @param in <Add Comment Here>
    * @see getPackage
    */
    public void setPackage(String in)
    {
        m_Package = in;
    }

    /**
    * code to get parameter AddedImports
    * @return <Add Comment Here>
    * @see getAddedImports
    */
    public String getAddedImports()
    {
        return(m_AddedImports);
    }

    /**
    * code to set parameter AddedImports
    * @param in <Add Comment Here>
    * @see getAddedImports
    */
    public void setAddedImports(String in)
    {
        m_AddedImports = in;
    }

    public void buildCOMSurrogates(COMDispatch[] Items) throws IOException
    {
        m_DispatchClasses = new HashSet();
        for(int i = 0; i < Items.length; i++) {
            String TheName = Items[i].getName();
            if(TheName.startsWith("_"))
                m_DispatchClasses.add(TheName.substring(1));
            else
                 m_DispatchClasses.add(TheName);
       }
        for(int i = 0; i < Items.length; i++) 
            buildCOMSurrogate(Items[i]);
    }
    
    public boolean isDispatchClass(String Name)
    {
        return(m_DispatchClasses.contains(Name));
    }

    public void buildCOMSurrogate(COMDispatch TheItem) throws IOException
    {
        buildCOMSurrogateBase(TheItem);
        buildCOMSurrogateWrapper(TheItem);
    }
    
    protected void buildCOMSurrogateWrapper(COMDispatch TheItem) throws IOException
    {
        String ClassName = TheItem.getName();
        if(ClassName.startsWith("_"))
            ClassName = ClassName.substring(1);
        String FileName = getPackage() + "." +  ClassName;
        FileName = FileName.replace('.','/');
        FileName = FileName.replace('\\','/');
        FileName += ".java";
        if(getTargetDirectory() != null)
            FileName = getTargetDirectory().getAbsolutePath().replace('\\','/') + "/" + FileName;
        FileUtilities.guaranteeDirectory(FileName);
        File Test = new File(FileName);
        if(Test.exists())
            return; // do not overwrite
        PrintWriter out = new PrintWriter(new FileOutputStream(Test));
        buildCOMSurrogateWrapper(out,TheItem);
    }
        
    protected void buildCOMSurrogateBase(COMDispatch in) throws IOException
    {
        String ClassName = in.getName();
        if(!ClassName.startsWith("_"))
            ClassName = "_" + ClassName;
        String FileName = getPackage() + ".gen." +  ClassName;
        FileName = FileName.replace('.','/');
        FileName = FileName.replace('\\','/');
        FileName += ".java";
        if(getTargetDirectory() != null)
            FileName = getTargetDirectory().getAbsolutePath().replace('\\','/') + "/" + FileName;
        FileUtilities.guaranteeDirectory(FileName);
        File Test = new File(FileName);
        PrintWriter out = new PrintWriter(new FileOutputStream(Test));
        buildCOMSurrogate(out,in);
    }
    
    protected void buildCOMSurrogateWrapper(PrintWriter out,COMDispatch in) throws IOException
    {
        // Some classes start with "_" - we change those names
        String ClassName = in.getName();
        if(ClassName.startsWith("_"))
            ClassName = in.getName().substring(1);
        writeWrapperHeader(ClassName,out,in);
        writeWrapperConstructors(ClassName,out,in);
        writeMethods(out,in,true);
        writeFooter(ClassName,out,in);
        out.close();
   }
    
    protected void buildCOMSurrogate(PrintWriter out,COMDispatch in) throws IOException
    {
        // Some classes start with "_" - we change those names
        String ClassName = in.getName();
        if(!ClassName.startsWith("_"))
            ClassName = "_" + in.getName();
        writeHeader(ClassName,out,in);
        writeConstructors(ClassName,out,in);
        writeMethods(out,in,false);
        writeFooter(ClassName,out,in);
        out.close();
    }
    
    protected void initMethodsParamHash(PrintWriter out,COMDispatch in) throws IOException
    {
        if(!writeMethodsParamHash(out,in))
            return; // nothing to do
        writeStaticInitializer(out,in,1);
    }
    
    protected void writeStaticInitializer(PrintWriter out,COMDispatch in,int indent) throws IOException
    {
        COMMethod[] methods = in.getMethods();
        Util.indent(out,indent);
        out.println("// Static initializer");
        Util.indent(out,indent);
        out.println("// Used to initialize Hash Tables for named parameters");
        Util.indent(out,indent);
        out.println("static {");
        for(int i = 0; i < methods.length; i++) {
            int NOptional = methods[i].getNumberOptionalParams();
            if(NOptional > NUM_NAMED_OPTIONAL_PARAMS) {
                writeMethodStaticInitializer(out,methods[i],indent + 1);
            }
        }
        Util.indent(out,indent);
        out.println("}");
    }
    
    protected void writeMethodStaticInitializer(PrintWriter out,COMMethod in,int indent) throws IOException
    {
        String HashName = buildHashName(in);
        Util.indent(out,indent);
        out.println(HashName + " = new HashMap();");
        COMParam[] TheParams = in.getParams();
        for(int i = 0; i < TheParams.length; i++) 
            writeParamHashItem(out,TheParams[i],HashName,i,indent);
        out.println();
    }
    
    protected void writeParamHashItem(PrintWriter out,COMParam in,String HashName,int index,int indent) throws IOException
    {
        Util.indent(out,indent);
        out.print(HashName + ".put(");
        out.print("\"" + in.getName() + "\",new Integer(" + index + ")");
        out.println(");");
    }
    
    protected boolean writeMethodsParamHash(PrintWriter out,COMDispatch in) throws IOException
    {
        COMMethod[] methods = in.getMethods();
        boolean hasHashMethod = false;
        for(int i = 0; i < methods.length; i++) 
            hasHashMethod |= writeMethodParamHash( out, methods[i]);
        return(hasHashMethod);
    }
    
    protected boolean writeMethodParamHash(PrintWriter out,COMMethod in) throws IOException
    {
        COMParam[] TheParams = in.getParams();
        int NOptional = in.getNumberOptionalParams();
        if(NOptional > NUM_NAMED_OPTIONAL_PARAMS) {
            Util.indent(out,1);
            out.println("protected static Map " + buildHashName(in) + ";");
            return(true);
        }
        return(false);
    }
    
    protected void writeMethods(PrintWriter out,COMDispatch in,boolean isWrapper) throws IOException
    {
        COMMethod[] methods = in.getMethods();
        for(int i = 0; i < methods.length; i++) 
            writeMethod( out, methods[i],isWrapper);
    }
    
    protected void writeMethod(PrintWriter out,COMMethod in,boolean isWrapper) throws IOException
    {
        COMParam[] TheParams = in.getParams();
        writeMethod(out,in,TheParams,isWrapper);
        int NOptional = in.getNumberOptionalParams();
        if(NOptional > NUM_NAMED_OPTIONAL_PARAMS)
            writeNamedParamCall(out,in,TheParams,isWrapper);
        while(NOptional > 0) {
            COMParam[] newParams = buildReducedParamList(TheParams,--NOptional);
            writeMethod(out,in,newParams,isWrapper);
        }
    }
    
    /**
    * bui;d a list of parameters with only n parametera
    */
    protected COMParam[] buildReducedParamList(COMParam[] TheParams,int n)
    {
        List holder = new ArrayList();
        for(int i = 0; i < TheParams.length; i++) {
            if(TheParams[i].isOptional()) {
                if(n > 0) {
                    holder.add(TheParams[i]);
                    n--;
                }
            }
            else {
                holder.add(TheParams[i]);
            }
        }
            
        COMParam[] ret = new COMParam[holder.size()];
        holder.toArray(ret);
        return(ret);
        
    }
        
    protected void writeMethod(PrintWriter out,COMMethod in,COMParam[] TheParams,boolean isWrapper) throws IOException
    {
        boolean returnAmbiguous = in.isReturnAmbiguous();
        if(isWrapper && !returnAmbiguous)
            return; // only gen wrapper methods for ambiguous returns
        boolean isBadReturn = in.hasReturn() && in.getReturn().getType().startsWith("@@@");
        int indent = 1;
        if(isWrapper)
            writeWrapperComment(out,in,TheParams,indent);
        else
            writeComment(out,in,TheParams,indent);
        // Comment out the code - we cannot handle this 
        // type
        if(isBadReturn) {
            Util.indent(out,indent);
            out.println("/*");
        }
        Util.indent(out,indent);
        out.print("public " + buildMethodReturn(in)+ " ");
        
        // make the name nonConflicting so we can put a real return
        // in the derived class
       if(returnAmbiguous && !isWrapper)
            out.print("ambig_"); 
            
        out.print(buildMethodName(in) + "(");
        writeMethodCallParams(out,TheParams);
        out.println(")");
        Util.indent(out,indent);
        out.println("{");
        if(isWrapper)
            writeWrapperMethodBody(out,in,TheParams);
        else
            writeMethodBody(out,in,TheParams);
        Util.indent(out,indent);
        out.println("}");
        // Comment out the code - we cannot handle this 
        // type
        if(isBadReturn) {
            Util.indent(out,indent);
            out.println("*/");
        }
        out.println();
    }
    
    protected static boolean hasBadParam(COMParam[] TheParams)
    {
        for(int i = 0; i < TheParams.length; i++) {
            if(TheParams[i].getType().startsWith("@@@"))
                return(true);
        }
        return(false);
    }
    
    protected void writeCommentBody(PrintWriter out,COMMethod in,COMParam[] TheParams,int indent) throws IOException
    {
        if(hasBadParam(TheParams)) {
            Util.indent(out,indent);
            out.print("* Undefined Param Type - Edit the XML Source ");
        }
            
        for(int i = 0; i < TheParams.length; i++) {
            writeParamTag(out,TheParams[i],indent,i);
        }
        for(int i = 0; i < TheParams.length; i++) {
            writeParamTag(out,TheParams[i],indent,i);
        }
        if(in.hasReturn()) {
            Util.indent(out,indent);
            out.println("* @return " + buildMethodReturn(in));
        }
            
    }
    
    protected void writeParamTag(PrintWriter out,COMParam param,int indent,int i) throws IOException
    {
        Util.indent(out,indent);
        out.print("* @param ");
        out.print(buildMethodParamCallName(param,i));
        String Type = buildMethodParamName(param);
        out.print(Type);
        if(param.isTypeEnum() && Type.equals("int"))
            out.print("RealType - " + param.getCom_Type().substring(COMParam.USER_DEF_PREFIX.length()));
           
        out.println();
    }
    
    protected void writeComment(PrintWriter out,COMMethod in,COMParam[] TheParams,int indent) throws IOException
    {
        Util.indent(out,indent);
        out.println("/**");
        writeCommentBody(out,in,TheParams,indent);
        Util.indent(out,indent);
        out.println("*/");
    }
    
    protected void writeWrapperComment(PrintWriter out,COMMethod in,COMParam[] TheParams,int indent) throws IOException
    {
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* This method is provided to allow the developer ");
         Util.indent(out,indent);
        out.println("* To reset an ambiguous return type -  ");
         Util.indent(out,indent);
        out.println("* It is anticipated the return will be changed  ");
        writeCommentBody(out,in,TheParams,indent);
        Util.indent(out,indent);
        out.println("*/");
    }
    
    protected void writeHashComment(PrintWriter out,COMMethod in,COMParam[] TheParams,int indent) throws IOException
    {
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* This call supports calling with named arguments as does VB");
        Util.indent(out,indent);
        out.println("* @param TheParams non-null Map where key is ");
        Util.indent(out,indent);
        out.println("*    param name and value is param value");
        if(in.hasReturn()) {
            Util.indent(out,indent);
            out.println("* @return " + buildMethodReturn(in));
        }
        Util.indent(out,indent);
        out.println("*/");
    }
    protected void writeHashWrapperComment(PrintWriter out,COMMethod in,COMParam[] TheParams,int indent) throws IOException
    {
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* This method is provided to allow the developer ");
        Util.indent(out,indent);
        out.println("* To reset an ambiguous return type -  ");
        Util.indent(out,indent);
        out.println("* It is anticipated the return will be changed  ");
        Util.indent(out,indent);
        out.println("* This call supports calling with named arguments as does VB");
        Util.indent(out,indent);
        out.println("* @param TheParams non-null Map where key is ");
        Util.indent(out,indent);
        out.println("*    param name and value is param value");
        if(in.hasReturn()) {
            Util.indent(out,indent);
            out.println("* @return " + buildMethodReturn(in));
        }
        Util.indent(out,indent);
        out.println("*/");
    }
    
    protected void writeWrapperMethodBody(PrintWriter out,COMMethod in,COMParam[] TheParams) throws IOException
    {
        int indent = 2;
        
        Util.indent(out,indent);
        out.print("return(ambig_" + buildMethodName(in) + "(");
        out.print(buildReCallArgumentString(TheParams,indent));
        out.println("));");
    }
    
    protected void writeMethodBody(PrintWriter out,COMMethod in,COMParam[] TheParams) throws IOException
    {
        int indent = 2;
        
        Util.indent(out,indent);
        out.print("Variant[] args = ");
        out.println(buildCallArgumentString(TheParams,indent));
        writeInvokeCall(out,in,indent);
    }
        
    protected void writeInvokeCall(PrintWriter out,COMMethod in,int indent) throws IOException
    {
        if(in.hasReturn()) {
            Util.indent(out,indent);
            out.print("Variant retVar = invoke_method(args,");
            out.print(in.getDispID());
            out.print(",");
            out.print(buildFlags(in));
            out.println(");");
            Util.indent(out,indent);
            out.println("if(retVar == null)");
            Util.indent(out,indent + 1);
            out.print("return(");
            out.print(buildNullReturnString(in.getReturn()));
            out.println(");");
            Util.indent(out,indent);
            out.println(buildReturnConvertString(in.getReturn()));
            Util.indent(out,indent);
            out.println("return(ret);");
        }
        else {
            Util.indent(out,indent);
            out.print("invoke_method_void(args,");
            out.print(in.getDispID());
            out.print(",");
            out.print(buildFlags(in));
            out.println(");");
        }
            
    }
    
    protected String buildNullReturnString(COMParam in)
    {
        if(in == null)
            return("null");
        if(in.isTypeEnum()) {
            String test = buildEnumArgument(in);
            if(test.equals("int"))
                return("0");
            return("null");
        }
        else {
            String Type = in.getType();
            if(Type.equals("int"))
                return("0");
            if(Type.equals("boolean"))
                return("false");
            if(Type.equals("float"))
                return("0");
            if (Type.equals("double")) 
                return("0");
            if(Type.equals("char"))
                return("0");
            if(Type.equals("short"))
                return("0");
            if(Type.equals("long"))
                return("0");
            return("null");
        }
    }

    
    protected String buildFlags(COMMethod in)
    {
        String Kind = in.getKind();
        Object ret = gKindToDispatchConstant.get(Kind);
        if(ret == null)
            throw new IllegalArgumentException("Kind " + Kind + " is not understood");
        return(ret.toString());
    }
    
    protected String buildReCallArgumentString(COMParam[] TheParams,int indent) 
    {
        if(TheParams.length == 0)
            return("");
        
        StringBuffer holder = new StringBuffer();
        for(int i = 0; i < TheParams.length; i++) {
            if(i != 0)
                holder.append(",");
            holder.append(buildMethodParamCallName(TheParams[i],i));
        }
         return(holder.toString());
    }
    
    protected String buildCallArgumentString(COMParam[] TheParams,int indent) 
    {
        if(TheParams.length == 0)
            return("Variant.EMPTY_ARRAY;");
        
        StringBuffer holder = new StringBuffer();
        holder.append("{");
        for(int i = 0; i < TheParams.length; i++) 
            holder.append(buildConvertParamToVariant(TheParams[i],i));
        holder.append("};");
        return(holder.toString());
    }
    
    protected String buildReturnConvertString(COMParam in) 
    {
        StringBuffer holder = new StringBuffer();
        String Type = buildMethodParamName(in);
        if(Type.equals("IDispatch") )
            Type = "Dispatch";
        if(Type.equals("IUnknown") )
            return("IUnknown ret = new IUnknown(retVar.toInt());");
        // Return Application not _Application
        if(Type.startsWith("_"))
            Type = Type.substring(1);
            
        holder.append(Type);
        holder.append(" ret = ");
        if(isDispatchClass(Type)) {
            holder.append("new " + Type + "(retVar.toDispatch())");
        }
        else {    
            // Handle enum substitution
            if(in.isTypeEnum() && !Type.equals("int")) {
                holder.append(Type + ".getOptionByValue(retVar.toInt())");
            }
            else {
                Object Converter = gTypeToVariantOp.get(Type);
                    if(Converter != null)
                        holder.append("retVar." + Converter + "()");
                    else {
                        if(Type.equals("java.util.Date")) {
                            holder.append("COMUtilities.fromVBDate(retVar.toDate()).getTime()");
                        }
                        else {
                            if(Type.equals("Dispatch") || Type.equals("Object"))
                                holder.append("retVar.ForceObjectVal()");
                            else   
                                holder.append("new " + Type + "((Dispatch)retVar.ForceObjectVal())");
                        }
              }
            }
        }
        holder.append(";");
        return(holder.toString());
    }
    
    
    protected String buildOneCallArgumentString(COMParam in,int i) 
    {
        StringBuffer holder = new StringBuffer(",");
        String ArgName = buildMethodParamCallName(in,i);
        holder.append(buildConvertParamToObject(in,ArgName));
        return(holder.toString());
    }
    
    protected String buildConvertParamToVariant(COMParam in,int i) 
    {
        StringBuffer holder = new StringBuffer();
        if(i != 0)
            holder.append(",");
        String ArgName = buildMethodParamCallName(in,i);
        // Handle enum substitution
        if(in.isTypeEnum() && !buildEnumArgument(in).equals("int")) {
            holder.append("new Variant(" + ArgName + ".getValue())");
            return(holder.toString());
        }
        // Handle enum substitution
        if(in.getType().equals("java.util.Date")) {
            holder.append("new Variant(COMUtilities.vbDate(" + ArgName + "))");
        }
        else {
            holder.append("new Variant(" + ArgName + ")");
        }
        return(holder.toString());
    }
    
    
    protected String buildConvertParamToObject(COMParam in,String ArgName) 
    {
        String Type = in.getType();
        if(ClassUtilities.isPrimitiveType(Type)) {
            return("new " + ClassUtilities.classNameToWrapperName(in.getType()) + "(" + ArgName + ")");
        }
        else {
            return(ArgName);
        }
            
    }
    
    protected void writeMethodCallParams(PrintWriter out,COMParam[] TheParams) throws IOException
    {
        for(int i = 0; i < TheParams.length; i++) 
            writeMethodCallParam(out,TheParams[i],i,i < (TheParams.length - 1));
    }
    
    protected void writeMethodCallParam(PrintWriter out,COMParam in,int i,boolean addComma) throws IOException
    {
         
        out.print(buildMethodParamName(in));
        out.print(" ");
        out.print(buildMethodParamCallName(in,i));
        if(addComma)
            out.print(",");
    }
    
    protected static String buildMethodName(COMMethod in)
    {
        String ret = unsafeBuildMethodName(in);
        if(ret.equals("getClass"))
            return("getClassX"); // we cannot override this name
        if(ret.equals("getDispatch"))
            return("getDispatchX"); // we cannot override this name
        if(ret.equals("getClassName"))
            return("getClassNameX"); // we cannot override this name
        return(ret);
    }
    
    
    protected static String unsafeBuildMethodName(COMMethod in)
    {
        String Kind = in.getKind();
        if(Kind.equals("IDispatch.PROPERTYGET"))
            return("get" + in.getName());
        if(Kind.equals("IDispatch.PROPERTYPUT"))
            return("set" + in.getName());
        if(Kind.equals("IDispatch.PROPERTYPUTREF"))
            return("set" + in.getName());
        if(Kind.equals("IDispatch.INVOKE_PROPERTYPUTREF"))
            return("set" + in.getName());
        if(Kind.equals("INVOKE_PROPERTYPUTREF"))
            return("set" + in.getName());
        if(Kind.equals("IDispatch.METHOD"))
            return(in.getName());
        throw new IllegalArgumentException("Uknown type " + Kind);
    }
    
    protected String buildMethodReturn(COMMethod in)
    {
        if(in.hasReturn()) {
            String ret = buildMethodParamName(in.getReturn());
            if(ret.startsWith("_"))
                ret = ret.substring(1);
            return(ret);
        }
        else {
            return("void");
        }
    }
    
    
    protected static String buildMethodParamCallName(COMParam in,int i )
    {
        if(in.getName().equals("unknown"))
            return("arg" + (i + 1));
        else
            return(in.getName());
    }
    
    protected String buildMethodParamName(COMParam in)
    {
        if(in.isTypeEnum())
            return(buildEnumArgument(in));
        String Type = in.getType();
        if(Type.equals("IDispatch"))
            return("Dispatch");
        if(Type.startsWith("@@@"))
            return("Object // " + Type + " \n");
            
        if(Type.endsWith("Exception")) { // Microsoft likes to use this name
            if(Type.equals("_Exception")) // Microsoft likes to use this name
                Type = Type.substring(1);
            if(Type.equals("Exception")) // Microsoft likes to use this name
                Type = getPackage() + "." + Type;
        }
        if(Type.endsWith("Error")) { // Microsoft likes to use this name
            if(Type.equals("_Error")) // Microsoft likes to use this name
                Type = Type.substring(1);
            if(Type.equals("Error")) // Microsoft likes to use this name
                Type = getPackage() + "." + Type;
        }
        if(Type.endsWith("System")) { // Microsoft likes to use this name
            if(Type.equals("_System")) // Microsoft likes to use this name
                Type = Type.substring(1);
            if(Type.equals("System")) // Microsoft likes to use this name
                Type = getPackage() + "." + Type;
        }
/*        if(Type.endsWith("System")) {
          Type=getPackage()+"."+Type; // Microsoft likes to use this name
        }*/
        return(Type);
    }
    
    protected String buildEnumArgument(COMParam in)
    {
        String EnumType = in.getCom_Type().substring(COMParam.USER_DEF_PREFIX.length());
        if(getTypeLib().findEnum(EnumType) != null)
            return(EnumType + TypelibEnumBuilder.ENUM_SUFFIX);
        return("int"); // we cannot handle the enum and this must be an int
    }
    
    protected void writeWrapperConstructors(String ClassName,PrintWriter out,COMDispatch in)
    {
        int indent = 1;
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* default constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "()");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent);
        out.println("}");
        
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* populating constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "(Dispatch  " + " in)");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("super(in);");
        Util.indent(out,indent);
        out.println("}");
        
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* populating constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "(WDispatch in)");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("super(in);");
        Util.indent(out,indent + 1);
        out.println("if(!getClsId().equals(in.getClsId())) {");
        Util.indent(out,indent + 2);
        out.println("if(!COMUtilities.allowCOMCast(getClass(),in.getClass()))");
        Util.indent(out,indent + 3);
        out.println("throw new IllegalArgumentException(\"input cannot be assigned\");");
        Util.indent(out,indent + 1);
        out.println("}");
        Util.indent(out,indent);
        out.println("}");
   }
    
    protected void writeConstructors(String ClassName,PrintWriter out,COMDispatch in)
    {
        int indent = 1;
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* default constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "()");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("super(clsid);");
        Util.indent(out,indent);
        out.println("}");
        
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* default constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "(Dispatch in)");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("super(in);");
        Util.indent(out,indent);
        out.println("}");
        
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* default constructor");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public " + ClassName + "(WDispatch in)");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("super(in);");
        Util.indent(out,indent + 1);
        out.println("if(!getClsId().equals(in.getClsId())) {");
        Util.indent(out,indent + 2);
        out.println("if(!COMUtilities.allowCOMCast(getClass(),in.getClass()))");
        Util.indent(out,indent + 3);
        out.println("throw new IllegalArgumentException(\"input cannot be assigned\");");
        Util.indent(out,indent + 1);
        out.println("}");
        Util.indent(out,indent);
        out.println("}");
         
        Util.indent(out,indent);
        out.println("/**");
        Util.indent(out,indent);
        out.println("* class id");
        Util.indent(out,indent);
        out.println("*/");
        Util.indent(out,indent);
        out.println("public String getClsId()");
        Util.indent(out,indent);
        out.println("{");
        Util.indent(out,indent + 1);
        out.println("return(clsid);");
        Util.indent(out,indent);
        out.println("}");
    }
   
    
    protected void writeWrapperHeader(String ClassName,PrintWriter out,COMDispatch in) throws IOException
    {
        out.println("package " + getPackage() + ";");
        out.println("import " + getPackage() + ".gen.*;");
        out.println("import " + getPackage() + ".com_enum.*;");
        out.println("import " + ClassUtilities.packageName(in) + ".*;");
        out.println("import java.util.Map;" );
        out.println("import java.util.Iterator;" );
        out.println("import java.util.HashMap;" );
        out.println("import com.jacob.com.*;" );
         if(getAddedImports().length() > 0)
            out.println(getAddedImports());
        out.println("import com.lordjoe.utilities.COMUtilities;" );

        
        out.println();
        out.println("/**");
        out.println("* java surrogate for com object " + in.getName());
        out.println("* It is OK to edit this code ");
        out.println("* @author S Lewis");
        out.println("*/");
        out.println("public class " + ClassName + " extends _" + ClassName);
        out.println("{");
        
    }
    
    protected void writeHeader(String ClassName,PrintWriter out,COMDispatch in) throws IOException
    {
        out.println("package " + getPackage() + ".gen;");
        out.println("import " + getPackage() + ".*;");
        out.println("import " + getPackage() + ".com_enum.*;");
        out.println("import " + ClassUtilities.packageName(in) + ".*;");
        out.println("import java.util.HashMap;" );
        out.println("import java.io.*;" );
        out.println("import java.util.Map;" );
        out.println("import java.util.Iterator;" );
        out.println("import com.jacob.com.*;" );
        out.println("import com.lordjoe.utilities.COMUtilities;" );
        
        out.println();
        out.println("/**");
        out.println("* java surrogate for com object " + in.getName());
        out.println("* DO NOT EDIT THIS CODE ");
        out.println("* EDIT the subclass " + ClassName.substring((1)));
        out.println("* @author S Lewis");
        out.println("*/");
        out.println("public class " + ClassName + " extends WDispatch");
        out.println("{");
        Util.indent(out,1);
        String IID = in.getIID();
        if(in.getName().endsWith("Application"))
            IID = in.getIID(); // break here
        COMCOClass CoClass = getTypeLib().findCOClass(in.getName());
        if(CoClass != null)
            IID = CoClass.getIID();
        out.println("public static final String clsid = \"" + IID + "\";");
        
        initMethodsParamHash(out,in);
    }
    protected void writeFooter(String Prefix,PrintWriter out,COMDispatch in) throws IOException
    {
        out.println("}");
        
    }


    protected void writeNamedParamCall(PrintWriter out,COMMethod in,COMParam[] TheParams,boolean isWrapper) throws IOException
    {
        boolean returnAmbiguous = in.isReturnAmbiguous();
        if(isWrapper && !returnAmbiguous)
            return; // only gen wrapper methods for ambiguous returns
        boolean isBadReturn = in.hasReturn() && in.getReturn().getType().startsWith("@@@");
        int indent = 1;
        if(isWrapper)
            writeHashWrapperComment(out,in,TheParams,indent);
        else
            writeHashComment(out,in,TheParams,indent);
            
        // Comment out the code - we cannot handle this 
        // type
   //     if(isBadReturn) {
    //        Util.indent(out,indent);
    //        out.println("/*");
    //    }


        Util.indent(out,indent);
        out.print("public " + buildMethodReturn(in)+ " ");
        
        // make the name nonConflicting so we can put a real return
        // in the derived class
       if(returnAmbiguous && !isWrapper)
            out.print("ambig_"); 
            
        out.println(buildMethodName(in) + "(Map TheParams)");
        Util.indent(out,indent);
        out.println("{");
        if(isWrapper)
            writeWrapperHashMethodBody(out,in,TheParams);
        else
            writeHashMethodBody(out,in,TheParams);
        Util.indent(out,indent);
        out.println("}");
        // Comment out the code - we cannot handle this 
        // type
      //  if(isBadReturn) {
      //      Util.indent(out,indent);
      //      out.println("*/");
      //  }
        out.println();
    }
    protected void writeWrapperHashMethodBody(PrintWriter out,COMMethod in,COMParam[] TheParams) throws IOException
    {
        int indent = 2;
        
        Util.indent(out,indent);
        out.print("return(ambig_" + buildMethodName(in) + "(TheParams)");
        out.println(");");
    }

    protected void writeHashMethodBody(PrintWriter out,COMMethod in,COMParam[] TheParams) throws IOException
    {
        int indent = 2;
        String HashName = buildHashName(in);
        
        Util.indent(out,indent);
        out.print("Variant[] args = {");
        out.print("new Variant(null)");
        for(int i = 1; i < TheParams.length; i++) {
            out.print(", new Variant(null)");
        }
        out.println("};");
        
        Util.indent(out,indent);
        out.println("Iterator it = TheParams.keySet().iterator();");
        Util.indent(out,indent);
        out.println("while(it.hasNext()) {");
        Util.indent(out,indent + 1);
        out.println("String ParamName = (String)it.next();" );
        Util.indent(out,indent + 1);
        out.println("Integer Item = (Integer)" + HashName + ".get(ParamName);" );
        Util.indent(out,indent + 1);
        out.println("Variant RealItem = new Variant(TheParams.get(ParamName));");
        Util.indent(out,indent + 1);
        out.println("args[Item.intValue()] = RealItem;" );
        Util.indent(out,indent);
        out.println("}");
        
        writeInvokeCall(out,in,indent);
            
    }
    
    public static String buildHashName(COMMethod in)
    {
        return("gParamHash" + in.getName());
    }
}
