/**{ file
    @name Util.java
    @function this class - a Nullity implements a large number of
         useful functions
    @author> Steven M. Lewis
    @copyright>
	************************
	*  Copyright (c) 1996,97,98
	*  Steven M. Lewis
	*  www.LordJoe.com
	************************

  Updated to make it compile on Java 1.5
  Miika Nurminen
  
    @date> Fri 5. May 2006
    @version> 1.1
 
 

}*/
package com.lordjoe.utilities;
import com.lordjoe.lib.xml.*;
import java.io.*;
import java.util.*;
import java.util.Arrays;
import java.beans.*;
import java.lang.reflect.Array;
import java.text.*;
import java.net.*;
import java.lang.reflect.*;
/**{ class
    @name Util
    @function this class - a Nullity implements a large number of
         useful functions
}*/
abstract public class Util extends Nulleton {

    // ========================
    // Time Constants
    public static final int  ONE_SECOND = 1000;
    public static final int  ONE_MINUTE = 60*ONE_SECOND;
    public static final int  ONE_HOUR   = 60*ONE_MINUTE;
    public static final int ONE_DAY    = 24*ONE_HOUR;
    public static final int ONE_WEEK   = 7*ONE_DAY;

    // Probably should use subclasses whivh are immutable
    public static final Vector   EMPTY_VECTOR = new Vector();
    public static final Map      EMPTY_MAP = new HashMap();
    public static final List     EMPTY_LIST = new ArrayList();
    // ========================
    // Empty arrays - might as well be singletons to save
    // construction
    public static final String[]   EMPTY_STRING_ARRAY = new String[0];
    public static final Object[]   EMPTY_OBJECT_ARRAY = new Object[0];
    public static final Class[]    EMPTY_CLASS_ARRAY = new Class[0];
    public static final int[]      EMPTY_INT_ARRAY = new int[0];
    public static final double[]   EMPTY_DOUBLE_ARRAY = new double[0];
    public static final float[]    EMPTY_FLOAT_ARRAY = new float [0];
    public static final byte[]     EMPTY_BYTE_ARRAY = new byte[0];
    public static final char[]     EMPTY_CHAR_ARRAY = new char[0];
    public static final short[]    EMPTY_SHORT_ARRAY = new short[0];
    public static final boolean[]  EMPTY_BOOLEAN_ARRAY = new boolean[0];
    public static final Thread[]   EMPTY_THREAD_ARRAY = new Thread[0];
    public static final Comparator DOUBLE_COMPARE = new DoubleComparator();
    public static final Comparator STRING_COMPARE = new StringComparator();
    public static final Comparator INTEGER_COMPARE = new IntegerComparator();
    public static final Comparator DATE_COMPARE = new DateComparator();

    private static String[] gIndentStrings;
    private static String gLocalIPAddress;
    private static Map gEmptyArrays = new HashMap();
    
    public static final Random gRandomizer = new Random(System.currentTimeMillis());
    
    //- *******************
    //- Methods
    
    /**
    * return an empty array of the requested class
    * @param in non-null requested class
    * @return non-null empty array of that class
    */
    public static Object emptyArray(Class in)
    {
        synchronized(gEmptyArrays) {
            Object ret = gEmptyArrays.get(in);
            if(ret ==  null) {
                ret = Array.newInstance(in,0);
                gEmptyArrays.put(in,ret);
            }
            return(ret);
        }
    }
    /**
    * return an empty array of the requested class
    * @param in non-null requested class
    * @return non-null CLass of aht array
    */
    public static Class arrayClass(Class in)
    {
        return(emptyArray(in).getClass());
    }

    /**{ method
        @name copyInto
        @function copies all elements of src array into dst array up to the
             minimum common size
        @param src - source data
        @param dst - destination data
    }*/
    public static void copyInto(Object[] src,Object[] dst) {
        if(src != null && dst != null) {
            int n = Math.min(src.length,dst.length);
            System.arraycopy(src,0,dst,0,n);
        }
    }

    /**{ method
        @name copyInto
        @function copies all elements of src array into dst array up to the
             minimum common size
        @param src - source data
        @param dst - destination data
    }*/
    public static void copyInto(int[] src,int[] dst) {
        if(src != null && dst != null) {
            int n = Math.min(src.length,dst.length);
            System.arraycopy(src,0,dst,0,n);
        }
    }

    /**{ method
        @name copyInto
        @function copies all elements of src array into dst array up to the
             minimum common size
        @param src - source data
        @param dst - destination data
    }*/
    public static void copyInto(char[] src,char[] dst) {
        if(src != null && dst != null) {
            int n = Math.min(src.length,dst.length);
            System.arraycopy(src,0,dst,0,n);
        }
    }

    /**{ method
        @name copyInto
        @function copies all elements of src array into dst array up to the
             minimum common size
        @param src - source data
        @param dst - destination data
    }*/
    public static void copyInto(boolean[] src,boolean[] dst) {
        if(src != null && dst != null) {
            int n = Math.min(src.length,dst.length);
            System.arraycopy(src,0,dst,0,n);
        }
    }


// Parsing

    /**{ method
        @name parseTabDelimited
        @function break a StringBuffer containing 0 or more tabe into
             substrings separated by tabs - the tabs are dropped
        @param in the string
        @return collection of 0 or more substrings
    }*/
    public static String[] parseTabDelimited(StringBuffer in) {
        return(parseTabDelimited(new String(in)));
    }

    /**{ method
        @name parseTabDelimited
        @function break a char array containing 0 or more tabe into
             substrings separated by tabs - the tabs are dropped
        @UnusedParam> in the string
        @param OneLine the array
        @param count where to start
        @return collection of 0 or more substrings
    }*/
    public static String[] parseTabDelimited(char[] OneLine,int count) {
        String s = new String(OneLine,0,count);
        return(parseTabDelimited(s));
    }

    /**{ method
        @name parseTabDelimited
        @function break a String containing 0 or more tabe into
             substrings separated by tabs - the tabs are dropped
        @param s the string
        @return collection of 0 or more substrings
    }*/
    public static String[] parseTabDelimited(String s) {
        if(s == null || s.length() == 0) {
            return(new String[0]);
        }
        return(parseTokenDelimited(s,'\t'));
    }

    /**{ method
        @name parseCommaDelimited
        @function break a String containing 0 or more  into
             substrings separated by commas - the commas are dropped
        @param s the string
        @return collection of 0 or more substrings
    }*/
    public static String[] parseCommaDelimited(String s) {
        if(s == null || s.length() == 0) {
            return(new String[0]);
        }
        return(parseTokenDelimited(s,','));
    }

    /**{ method
        @name parseTokenDelimited
        @function break a String containing 0 or more occurances
             of the char token into
             substrings separated by token - the tokens are dropped
        @param s the string
        @param token delimiter token
        @return collection of 1 or more substrings
    }*/
    public static String[] parseTokenDelimited(String s,char token) {
        String[] ret;
        int SLength = s.length();
	if(SLength == 0)
	   return new String[0];
        // string does not hold token
        if(s.indexOf(token) == - 1) {
            ret = new String[1];
            ret[0] = s;
        }
        else {
            Vector tokens = new Vector();
            int LastToken = 0;
            int CurrentToken = s.indexOf(token,LastToken);
            while(CurrentToken != - 1) {
                String OneToken = s.substring(LastToken,CurrentToken);
                tokens.addElement(OneToken);
                LastToken = CurrentToken + 1;
                CurrentToken = s.indexOf(token,LastToken);
            }
            if(LastToken < SLength) {
                tokens.addElement(s.substring(LastToken,SLength));
            }
            ret = new String[tokens.size()];
            tokens.copyInto(ret);
        }
        return(ret);
    }

    /**{ method
        @name parseWhiteSpaceDelimited
        @function break a String containing 0 or more occurances
             of the char token into
             substrings separated by any white space
        @param s the string
        @return collection of 1 or more substrings
    }*/
    public static String[] parseWhiteSpaceDelimited(String s)
    {
        String[] ret;
        Vector tokens = new Vector();
        int SLength = s.length();
        StringBuffer accum = new StringBuffer(SLength);

        for(int i = 0; i < SLength; i++)
        {
            char c = s.charAt(i);
            if(c > ' ') {
                accum.append(c);
            }
            else { // white space
                if(accum.length() > 0) {
                    tokens.addElement(accum.toString());
                    accum.setLength(0);
                }
            }
        }
        if(accum.length() > 0) {
            tokens.addElement(accum.toString());
        }

        ret = new String[tokens.size()];
        tokens.copyInto(ret);
        return(ret);
    }

    /**{ method
        @name breakTokenDelimited
        @function break a String containing 0 or more occurances
             of the char token into
             substrings separated by token - the tokens are dropped
        @param s the string
        @param token delimiter token
        @return collection of 1 or 2 substrings
    }*/
    public static String[] breakTokenDelimited(String s,char token) {
        String[] ret;
        int CurrentToken = s.indexOf(token);
        // string does not hold token
        if(s.indexOf(token) == - 1) {
            ret = new String[1];
            ret[0] = s;
        }
        else {
            ret = new String[2];
            ret[0] = s.substring(0,CurrentToken);
            ret[1] = s.substring(CurrentToken + 1);
        }
        return(ret);
    }


    /**{ method
        @name parseLinesWithBlanks
        @function break a Stringbuffer to a collection of lines
             i.e. separate on \n
        @param s the buffer - must be non-empty
        @return the collection of lines
    }*/
    public static String[] parseLinesWithBlanks(String s) {
        List TheLines = new ArrayList();
        StringBuffer sb = new StringBuffer();
        int index = 0;
        while(index < s.length()) {
            char c = s.charAt(index++);
            if(c == '\r')
                continue;
            if(c == '\n') {
                TheLines.add(sb.toString());
                sb.setLength(0);
            }
            else {
                sb.append(c);
            }
        }
        if(sb.length() > 0)
            TheLines.add(sb.toString());
        String[] ret = new String[TheLines.size()];
        TheLines.toArray(ret);
        return(ret);
    }

    /**{ method
        @name parseLines
        @function break a Stringbuffer to a collection of lines
             i.e. separate on \n
        @param s the buffer - must be non-empty
        @return the collection of lines
    }*/
    public static String[] parseLines(String s) {
        List TheLines = new ArrayList();
        StringTokenizer t = new StringTokenizer(s,"\n\r");
        while(t.hasMoreTokens()) {
            String st = t.nextToken();
            TheLines.add(st);
        }
        String[] ret = new String[TheLines.size()];
        TheLines.toArray(ret);
        return(ret);
    }

    /**{ method
        @name parseLines
        @function break a Stringbuffer to a collection of lines
             i.e. separate on \n
        @param s the buffer - must be non-empty
        @return the collection of lines
    }*/
    public static String[] parseLines(StringBuffer s) {
        return(parseLines(s.toString()));
    }


    /**{ method
        @name getLine
        @function retrieve text till \n
        @param s text source
        @param buffer ???
        @param StartPoint where to start
        @return the line
    }*/
    protected static String getLine(StringBuffer s,IntegerRef StartPoint,StringBuffer buffer) {
        int start = StartPoint.value;
        buffer.setLength(0);
        while(start < s.length()) {
            char c = s.charAt(start ++);
            int n =(int) c;
            // peek at binary value for debugging
            if(c == '\n') {
                break;
            }
            // count printing chars and tab
            if(c >= ' ' || c == '\t') {
                buffer.append(c);
            }
            n = 0;
        }
        StartPoint.value = start;
        // drop trailing \r
        if(start < s.length()) {
            if(s.charAt(start) == '\r') {
                StartPoint.value ++;
            }
        }
        // ; is comment
        if(buffer.length() > 0 && buffer.charAt(0) != ';') {
            return(new String(buffer));
        }
        // valid line
        if(start >= s.length()) {
            return(null);
        }
        return(getLine(s,StartPoint,buffer));
    }

    /**{ method
        @name vectorToArray
        @function turn a vector into an array of objects
        @param v vector
        @return the array or null if v empty
    }*/
    public static Object[] vectorToArray(Vector v) {
       return(vectorToArray(v,java.lang.Object.class));
    }

    /**{ method
        @name vectorToArray
        @function turn a vector into an array of objects
        @param v vector
        @param TargetClass class used by the array
        @return a non-null array
    }*/
    public static Object[] vectorToArray(Vector v,Class TargetClass) {
        return(collectionToArray(v,TargetClass));
    }

    /**{ method
        @name hashtableKeysToArray
        @function turn a Hashtable keys into an array of objects
        @param v Thetable - non-null
        @param TargetClass class for the array - non-null
        @return and array ot objects ot type TargetClass = non-null
    }*/
    public static Object[] hashtableKeysToArray(Hashtable v,Class TargetClass) {
        Enumeration e = v.keys();
        int count = 0;
        Object[] out = (Object[])java.lang.reflect.Array.newInstance(TargetClass, v.size());
        while(e.hasMoreElements()) {
            out[count++] = e.nextElement();
        }
        return(out);
    }
    /**{ method
        @name hashtableValuesToArray
        @function turn a Hashtable values into an array of objects
        @param v Thetable - non-null
        @param TargetClass class for the array - non-null
        @return and array ot objects ot type TargetClass = non-null
    }*/
    public static Object[] hashtableValuesToArray(Hashtable v,Class TargetClass) {
        Enumeration e = v.elements();
        int count = 0;
        Object[] out = (Object[])java.lang.reflect.Array.newInstance(TargetClass, v.size());
        while(e.hasMoreElements()) {
            out[count++] = e.nextElement();
        }
        return(out);
    }

    /**{ method
        @name hashtableFillsArray
        @function turn a Hashtable into an array of objects
        @param v Thetable
        @param out Array of appropriate type sized to v.size
        @return the array or null if v empty
    }*/
    public static Object[] hashtableFillsArray(Hashtable v,Object[] out) {
        Enumeration e = v.keys();
        int count = 0;
        Object added; // Better debugging on failure
        try {
            while(e.hasMoreElements()) {
                added = e.nextElement();
                out[count++] = added;
            }
        }
        catch(ArrayStoreException ex) {
            Assertion.fatalException(ex);
        }
        return(out);
    }
    
    

    /**{ method
        @name numberLeadingBlanks
        @function return number of blanks at the head of a string
        @param s non-null test string
        @return the number
        @policy rarely override
        @primary
    }*/
    public int numberLeadingBlanks(String s) {
        int n = 0;
        for(int i = 0; i < s.length(); i ++) {
            if(s.charAt(i) != ' ') {
                return(n);
            }
            else {
                n ++;
            }
        }
        return(n);
    }

    /**{ method
        @name compareStringArrays
        @function test is two string arrays are equal
        @param data array 1 to test - null or 0 length OK
        @param test array 2 to test - null or 0 length OK
        @return true if equal
    }*/
    public static boolean compareStringArrays(String[] data,String[] test) {
        if(data == null) {
            return(test == null);
        }
        if(test == null) {
            return(false);
        }
        if(data.length != test.length) {
            return(false);
        }
        for(int i = 0; i < data.length; i ++) {
            if(data[i] == null || ! data[i].equals(test[i])) {
                return(false);
            }
        }
        return(true);
        // perfect match
    }
    /**{ method
        @name stringsNotContained
        @function return an array of unique strings in base and not in drop
        @param base - non-null -  original array
        @param drop - non-null array of strings to drop
        @return  non-null array in the order of base
    }*/
    public static String[] stringsNotContained(String[] base,String[] drop)
    {
        Set inSet = arrayToSet(base);
        Set dropSet = arrayToSet(drop);
        inSet.removeAll(dropSet);
        List holder = new ArrayList(inSet.size());
        for(int i = 0; i < base.length; i++) {
            if(inSet.contains(base[i])) {
                inSet.remove(base[i]);
                holder.add(base[i]);
            }
        }
        String[] ret = new String[holder.size()];
        holder.toArray(ret);
        return(ret);
    }

    /**{ method
        @name arrayToSet
        @function convert an 
        @param in - non-null arrsy to place
        @return  non-null set
    }*/
    public static Set arrayToSet(Object[] in)
    {
        Set ret = new HashSet();
        for(int i = 0; i < in.length; i++) 
            ret.add(in[i]);
        return(ret);
    }


    /**{ method
        @name uniqueSort
        @function return a sorted array of strings without duplicates
        @param base - non-null -  original array
        @return  non-null sorted set without duplicates
    }*/
    public static String[] uniqueSort(String[] base)
    {
        if(base.length == 0 )
            return(EMPTY_STRING_ARRAY);
        String[] BaseClone = Clone.clone(base);
        Vector holder = new Vector();
        sort(BaseClone);
        holder.addElement(BaseClone[0]);
        String Last = BaseClone[0];
        for(int i = 1; i < BaseClone.length; i++) {
            if(!Last.equals(BaseClone[i])) {
                Last = BaseClone[i];
                holder.addElement(BaseClone[i]);
            }
        }
        String[] ret = new String [holder.size()];
        holder.copyInto(ret);
        return(ret);

    }

    /**
    * Sort an array according to a property value
    * @param data non-null array of object 
    * @param PropName - name of a property of those objects
    * @return non-null array of same type and length as data but sorted
    */
    
    public static Object[] sortOnProperty(Object[] data,String PropName) {
        
        Class ArrayClass = data.getClass().getComponentType();
        ClassProperty TheProp = ClassAnalyzer.getClassProperty(ArrayClass, PropName);
        
        Object[] testData = (Object[])Array.newInstance(TheProp.m_GetMethod.getReturnType(),data.length);
        try {
            for(int i = 0; i < data.length; i++) {
                Object Added = TheProp.m_GetMethod.invoke(data[i],EMPTY_OBJECT_ARRAY);
                testData[i] = Added;
            }
        }
        catch(IllegalAccessException ex) {
            throw new IllegalArgumentException("IllegalAccessException getter for " + PropName + " not found");
        }
        catch(InvocationTargetException ex) {
            throw new IllegalArgumentException("InvokationTargetException getter for " + PropName + " not found");
        }
        int[] order = getSortOrder(testData);
        
        Object[] ret = Clone.clone(data);
        for(int i = 0; i < order.length; i++) 
            ret[i] = data[order[i]];
        return(ret);
    }

    /**{ method
        @name differenceStringArrays
        @function the test returns strings in base not in probe
        @param base - non-null - the test returns strings in base not in probe
        @param probe non-null probe array
        @return true if equal
    }*/
    public static String[] differenceStringArrays(String[] base,String[] probe) {
        String[] BaseClone = uniqueSort(base);
        String[] ProbeClone = uniqueSort(probe);
        Vector holder = new Vector();
         for(int probeIndex = 0,baseIndex = 0; probeIndex < ProbeClone.length; probeIndex++) {
            boolean TerminateSearch = false;
            while(!TerminateSearch) {
                int result = ProbeClone[probeIndex].compareTo( BaseClone[baseIndex]);
                if(result == 0) {
                    TerminateSearch = true; // break
                    baseIndex++;
                }
                else {
                    if(result > 0) {
                        baseIndex++;
                        if(baseIndex >= BaseClone.length) {
                            while(probeIndex < ProbeClone.length)
                                holder.addElement(ProbeClone[probeIndex++]); // not present
                            break;
                        }
                    }
                    else {
                        TerminateSearch = true; // break
                        holder.addElement(ProbeClone[probeIndex]); // not present
                    }
                }
            } // while(!TerminateSearch)
        }

        String[] ret = new String [holder.size()];
        holder.copyInto(ret);
        return(ret);
    }


    public static Object[] vectorToClassArray(Vector holder)
    {
        return(vectorToClassArray(holder,minimumClassInVector(holder)));
    }

    public static Object[] vectorToClassArray(Vector holder,Class ArrayClass)
    {
        Assertion.fixThis();
        return(null);
    }

    /**{ method
        @name minimumClassInVector
        @function find the class which all members are assignable from
        @param holder non-null vector to test
        @param test non-null object to test
        @return non-null class which all members are assignable from
    }*/
    public static Class minimumClassInVector(Vector holder)
    {
        if(holder.size() == 0)
            return(java.lang.Object.class);
        Class ret = holder.elementAt(0).getClass();
        for(int i = 1; i < holder.size(); i++)
            ret = minimumCommonClass(ret,holder.elementAt(i));
        return(ret);
    }



    /**{ method
        @name minimumCommonClass
        @function find the lowest level class in common with test and base
        @param base non-null class to test
        @param test non-null object to test
        @return non-null class which test is assignable from
    }*/
    public static Class minimumCommonClass(Class base,Object test)
    {
        if(base.isInstance(test))
            return(base);
        base = base.getSuperclass();
        if(base == null)
            return(java.lang.Object.class);
        return(minimumCommonClass(base,test));
    }


    /**{ method
        @name memberStringArray
        @function test whether a string is a member of a string array
        @param data array to test - null or 0 length OK
        @param test string to test - null OK
        @return true if a member
    }*/
    public static boolean memberStringArray(String[] data,String test) {
        return(positionStringArray(data,test) >= 0);
    }

    /**{ method
        @name positionStringArray
        @function test whether a string is a member of a string array and
             return its position
        @param data array to test - null or 0 length OK
        @param test string to test - null OK
        @return 0 or more if a member - -1 if not
    }*/
    public static int positionStringArray(String[] data,String test) {
        if(data == null || data.length == 0) {
            return(- 1);
            // not found - no data
        }
        return(positionStringArray(data,test,0,data.length));
        // search whole array
    }

    /**{ method
        @name positionStringArray
        @function test whether a string is a member of a string array and
             return its position
        @param data array to test - null or 0 length OK
        @param test string to test - null OK
        @param start <Add Comment Here>
        @param end <Add Comment Here>
        @return 0 or more if a member - -1 if not
    }*/
    public static int positionStringArray(String[] data,String test,int start,int end) {
        if(data == null || data.length == 0) {
            return(- 1);
            // not found
        }
        if(test == null) {
            return(- 1);
            // not found
        }
        for(int i = start; i < end; i ++) {
            if(data[i] == null || data[i].equals(test)) {
                return(i);
            }
            // match
        }
        return(- 1);
        // not found
    }

    /**{ method
        @name readBytes
        @function copy bytes from one String buffer into another at the end
        @param src source buffer
        @param dst destination buffer
        @param size number bytes to copy
        @param Current holds start byte at end holds last copied + 1
    }*/
    public static void readBytes(StringBuffer src,StringBuffer dst,int size,IntegerRef Current) {
        dst.setLength(0);
        // clear dst
        for(int i = 0; i < size; i ++) {
            dst.append(src.charAt(Current.value ++));
        }
    }

    /**{ method
        @name readBytes
        @function copy bytes from one String into a StringBuffer at the end
        @param src source buffer
        @param dst destination buffer
        @param size number bytes to copy
        @param Current holds start byte at end holds last copied + 1
    }*/
    public static void readBytes(String src,StringBuffer dst,int size,IntegerRef Current) {
        dst.setLength(0);
        // clear dst
        for(int i = 0; i < size; i ++) {
            dst.append(src.charAt(Current.value ++));
        }
    }

    /**{ method
        @name readString
        @function copy bytes from one StringBuffer a String - triming the result
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the string extracted
    }*/
    public static String readString(StringBuffer s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        return(dst.toString().trim());
    }

    /**{ method
        @name readStringNoTrim
        @function copy bytes from one StringBuffer into a String - not triming the result
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the string extracted
    }*/
    public static String readStringNoTrim(StringBuffer s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        return(dst.toString());
    }

    /**{ method
        @name readInt
        @function copy bytes from one StringBuffer into an int - the buffer is assumed to hold digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the int extracted
    }*/
    public static int readInt(StringBuffer s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        String temp = dst.toString();
        return(Integer.parseInt(temp));
    }

    /**{ method
        @name readInt
        @function copy bytes a String into an int - the buffer is assumed to hold digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the int extracted
    }*/
    public static int readInt(String s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        String temp = dst.toString().trim();
        if(temp.length() > 0) {
            return(Integer.parseInt(temp));
        }
        else {
            return(0);
        }
    }

    /**{ method
        @name readHex
        @function copy bytes from one StringBuffer into an int - the buffer is assumed to hold hex digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the int extracted
    }*/
    public static int readHex(StringBuffer s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        String temp = dst.toString();
        return(Integer.parseInt(temp,16));
    }

    /**{ method
        @name readHex
        @function copy bytes from a String into an int - the buffer is assumed to hold hex digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return the int extracted
    }*/
    public static int readHex(String s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        String temp = dst.toString().trim();
        if(temp.length() > 0) {
            return(Integer.parseInt(temp,16));
        }
        else {
            return(0);
        }
    }

    /**{ method
        @name readChars
        @function copy bytes from one StringBuffer into a StringBuffer -
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return a Stringbuffer holding the extracted characters
    }*/
    public static StringBuffer readChars(StringBuffer s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        char c;
        while(Current.value < s.length()) {
            c = s.charAt(Current.value ++);
            dst.append(c);
            if(dst.length() >= size) {
                break;
            }
        }
        return(dst);
    }

    /**{ method
        @name readChars
        @function copy bytes from a String into a StringBuffer -
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return a Stringbuffer holding the extracted characters
    }*/
    public static StringBuffer readChars(String s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        char c;
        while(Current.value < s.length()) {
            c = s.charAt(Current.value ++);
            dst.append(c);
            if(dst.length() >= size) {
                break;
            }
        }
        return(dst);
    }

    /**{ method
        @name readString
        @function copy bytes from one StringBuffer into an int - the buffer is assumed to hold hex digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return a String holding the extracted characters - trimmed
    }*/
    public static String readString(String s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        return(dst.toString().trim());
    }

    /**{ method
        @name readStringNoTrim
        @function copy bytes from one StringBuffer into an int - the buffer is assumed to hold hex digits
        @param s source buffer
        @param Current holds start byte at end holds last copied + 1
        @param size number bytes to copy
        @return a String holding the extracted characters - not trimmed
    }*/
    public static String readStringNoTrim(String s,int size,IntegerRef Current) {
        StringBuffer dst = new StringBuffer();
        readBytes(s,dst,size,Current);
        return(dst.toString());
    }

    /**{ method
        @name memberStringArray
        @function test if a string is in a string array
        @param data array
        @param test string to test
        @return true if test is in the array
    }*/
    public static boolean memberStringArray(String test,String[] data) {
        if(data == null || test == null) {
            return(false);
        }
        for(int i = 0; i < data.length; i ++) {
            if(test.equals(data[i])) {
                return(true);
            }
        }
        return(false);
        // not found
    }

    /**{ method
        @name toString
        @function convert a char to a string
        @param c the char
        @return the String
    }*/
    public static String toString(char c) {
        StringBuffer s = new StringBuffer();
        s.append(c);
        return(s.toString());
    }

    /**{ method
        @name caselessCompare
        @function Compares String s1 to another specified String s2.
             Returns an integer that is less than, equal to, or greater than zero.
             The integer's value depends on whether s1 is less than,
             equal to, or greater than s2. Unlike String.compareTo this
             ignores case
        @param s1 string1 cannot be null
        @param s2 string2 cannot be null
        @return the answer
    }*/
    public static int caselessCompare(String s1,String s2) {
        int l = Math.min(s1.length(),s2.length());
        for(int i = 0; i < l; i ++) {
            char c1 = Character.toUpperCase(s1.charAt(i));
            char c2 = Character.toUpperCase(s2.charAt(i));
            if(c1 != c2) {
                return(c1 - c2);
            }
        }
        if(s1.length() != s2.length()) {
            return(s1.length() - s2.length());
        }
        // all else fails compare with case
        return(s1.compareTo(s2));
    }

    /**{ method
        @name padStringWithBlanks
        @function add blanks to a String until it is of length N
        @param s the starting string
        @param n the required size
        @return resultant string
    }*/
    public static String padStringWithBlanks(String s,int n) {
        StringBuffer sb = new StringBuffer(s);
        if(sb.length() > n) {
            sb.setLength(n);
        }
        else {
            while(sb.length() < n)
                sb.append(' ');
        }
        return(sb.toString());
    }

    /**{ method
        @name filterControlCharacters
        @function Make a string without control characters
        @param in input string
        @return string without control characters
    }*/
    public static String filterControlCharacters(String in) {
        StringBuffer sb = new StringBuffer(in.length());
        for(int i = 0; i < in.length(); i ++) {
            char c = in.charAt(i);
            if(c >= ' ' && c < 127) {
                sb.append(c);
            }
            else {
                // tab = 4 spaces - other control chars ignored
                if(c == '\t') {
                    sb.append("   ");
                }
            }
        }
        return(sb.toString());
    }

    /**{ method
        @name nowPlus
        @function return the current date plus del microseconds
        @param del increment
        @return new date
    }*/
    public static GregorianCalendar nowPlus(long del) {
        GregorianCalendar now = new GregorianCalendar();
        now.setTime(new Date(now.getTime().getTime() + del));
        return(now);
    }

    /**{ method
        @name currentTime
        @function return the current time in millisec
        @return the time
    }*/
    public static long currentTime() {
        Date now = new Date();
        return(now.getTime());
    }

    /**{ method
        @name getReverseSortOrder
        @function <Add Comment Here>
        @param data <Add Comment Here>
        @return <Add Comment Here>
        @policy <Add Comment Here>
    }*/
    public static int[] getReverseSortOrder(Object[] data) {
        int[] ret = getSortOrder(data);
        reverse(ret);
        return(ret);
    }

    /**{ method
        @name getSortOrder
        @function <Add Comment Here>
        @param data <Add Comment Here>
        @return <Add Comment Here>
        @policy <Add Comment Here>
    }*/
    public static int[] getSortOrder(Object[] data) {
        int nrows = data.length;
        String[] realData = new String[nrows];
        int[] ra = new int[nrows];
        for(int k = 0; k < nrows; k ++) {
            ra[k] = k;
            realData[k] = data[k].toString();
        }
        indexedSort(realData,ra,STRING_COMPARE);
        return(ra);
    }
    
    /**{ method
        @name getSortOrder
        @function <Add Comment Here>
        @param data non-null Array to sort
        @param Comparison non-null comparison function
        @return non-null array of ints showing new sort order
        @policy <Add Comment Here>
    }*/
    public static int[] getSortOrder(Object[] data,Comparator Comparison) {
        int nrows = data.length;
        Object[] realData = new Object[nrows];
        int[] ra = new int[nrows];
        for(int k = 0; k < nrows; k ++) {
            ra[k] = k;
            realData[k] = data[k];
        }
        indexedSort(realData,ra,Comparison);
        return(ra);

    }
    private static int left_index(int i) {return 2*i+1;}
    private static int right_index(int i) {return 2*i+2;}
    /** Sort array, placing an array of integer indices into the
     *  array idx.  If idx is smaller than array, an
     *  IndexOutOfBoundsException will be thrown.
     *
     *  array[idx[i]] will be the i'th ranked element of array.
     *  e.g.  array[idx[0]] is the smallest element,
     *        array[idx[1]] is the next smallest and so on.
     *
     *  array is left_index in its original order.
     **/
    public static void indexedSort(final Object[] array,int idx[],Comparator Comparison) {
	    int sortsize=array.length;
	    int i,top,t,largest,l,r,here;
	    int temp;

	    if (sortsize<=1) {
	        if (sortsize==1) idx[0]=0;
	        return;
	    }
    	
	    top=sortsize-1;
	    t=sortsize/2;
    	
	    for (i=0;i<sortsize;++i) idx[i]=i;
    	
	    do {
	        --t;
	        largest=t;

	        /* heapify */

	        do {
		    i=largest;
		    l=left_index(largest);
		    r=right_index(largest);

		    if (l<=top) {
		        if (Comparison.compare(array[idx[l]],array[idx[i]])>0) largest=l;
		    }
		    if (r<=top) {
		        if (Comparison.compare(array[idx[r]],array[idx[largest]])>0) largest=r;
		    }
		    if (largest!=i) {
		        temp=idx[largest];
		        idx[largest]=idx[i];
		        idx[i]=temp;
		    }
	        }
	        while (largest!=i);
	    }
	    while (t>0);

	    t=sortsize;

	    do {
	        --top;
	        --t;

	        here=t;

	        temp=idx[here];
	        idx[here]=idx[0];
	        idx[0]=temp;

	        largest=0;

	        do {
		    i=largest;
		    l=left_index(largest);
		    r=right_index(largest);

		    if (l<=top) {
		        if (Comparison.compare(array[idx[l]],array[idx[i]])>0) largest=l;
		    }
		    if (r<=top) {
		        if (Comparison.compare(array[idx[r]],array[idx[largest]])>0) largest=r;
		    }
		    if (largest!=i) {
		        temp=idx[largest];
		        idx[largest]=idx[i];
		        idx[i]=temp;
		    }
	        }
	        while (largest!=i);
	    }
	    while (t>1);
    }

    /**{ method
        @name sort
        @function - alphabetize an array of strings in place
             using heapsort
        @param s strings to be alphabetized
    }*/
    public static void sort(String[] s) {
        if(s == null || s.length < 2) {
            return;
        }
        int l,j,ir,i;
        String rra;
        String[] ra = new String[s.length + 1];
        for(int ix = 0; ix < s.length; ix ++)
            ra[ix + 1] = s[ix];
        // now run the heap	sort
        l =(s.length >> 1) + 1;
        ir = s.length;
        for(;;) {
            if(l > 1) {
                rra = ra[-- l];
            }
            else {
                rra = ra[ir];
                ra[ir] = ra[1];
                if(-- ir == 1) {
                    ra[1] = rra;
                    break;
                }
            }
            i = l;
            j = l << 1;
            while(j <= ir) {
                if(j < ir && ra[j].compareTo(ra[j + 1]) < 0) {
                    ++ j;
                }
                if(rra.compareTo(ra[j]) < 0) {
                    ra[i] = ra[j];
                    j +=(i = j);
                }
                else {
                    j = ir + 1;
                }
            }
            ra[i] = rra;
        }
        for(int ix = 0; ix < s.length; ix ++)
            s[ix] = ra[ix + 1];
    }

    /**{ method
        @name objectSort
        @function - alphabetize an array of strings in place
             using heapsort
        @param s strings to be alphabetized
    }*/
    public static void objectSort(Object[] s) {
        if(s == null || s.length < 2) {
            return;
        }
        int l,j,ir,i;
        Object rra;
        Object[] ra = new Object[s.length + 1];
        for(int ix = 0; ix < s.length; ix ++)
            ra[ix + 1] = s[ix];
        // now run the heap	sort
        l =(s.length >> 1) + 1;
        ir = s.length;
        for(;;) {
            if(l > 1) {
                rra = ra[-- l];
            }
            else {
                rra = ra[ir];
                ra[ir] = ra[1];
                if(-- ir == 1) {
                    ra[1] = rra;
                    break;
                }
            }
            i = l;
            j = l << 1;
            while(j <= ir) {
                if(j < ir && ra[j].toString().compareTo(ra[j + 1].toString()) < 0) {
                    ++ j;
                }
                if(rra.toString().compareTo(ra[j].toString()) < 0) {
                    ra[i] = ra[j];
                    j +=(i = j);
                }
                else {
                    j = ir + 1;
                }
            }
            ra[i] = rra;
        }
        for(int ix = 0; ix < s.length; ix ++)
            s[ix] = ra[ix + 1];
    }

    /**{ method
        @name descendingSort
        @function - sort a collection of comparable objects
        @param s objects to be sorted
    }*/
    public static void descendingSort(Comparable[] s) {
        ascendingSort(s);
        reverse(s);
    }

    /**{ method
        @name ascendingSort
        @function - sort a collection of comparable objects
        @param s objects to be sorted
    }*/
    public static void ascendingSort(Comparable[] s) {
        if(s == null || s.length < 2) {
            return;
        }
        int l,j,ir,i;
        Comparable rra;
        Comparable[] ra = new Comparable[s.length + 1];
        for(int ix = 0; ix < s.length; ix ++)
            ra[ix + 1] = s[ix];
        // now run the heap	sort
        l =(s.length >> 1) + 1;
        ir = s.length;
        for(;;) {
            if(l > 1) {
                rra = ra[-- l];
            }
            else {
                rra = ra[ir];
                ra[ir] = ra[1];
                if(-- ir == 1) {
                    ra[1] = rra;
                    break;
                }
            }
            i = l;
            j = l << 1;
            while(j <= ir) {
                if(j < ir && ra[j].compareTo(ra[j + 1]) < 0) {
                    ++ j;
                }
                if(rra.compareTo(ra[j]) < 0) {
                    ra[i] = ra[j];
                    j +=(i = j);
                }
                else {
                    j = ir + 1;
                }
            }
            ra[i] = rra;
        }
        for(int ix = 0; ix < s.length; ix ++)
            s[ix] = ra[ix + 1];
    }

    /**{ method
        @name objectReverseSort
        @function - reverse alphabetize an array of strings in place
             using heapsort
        @param s strings to be alphabetized
    }*/
    public static void objectReverseSort(Object[] s) {
        objectSort(s);
        reverse(s);
    }

    /**{ method
        @name stringOfChar
        @function return a string containing n copies of the character c
        @param n - number blanks
        @param c - fil character
        @return the string
    }*/
    public static String stringOfChar(int n,char c) {
        StringBuffer ret = new StringBuffer(n);
        for(int i = 0; i < n; i++)
            ret.append(c);
        return(ret.toString());
    }
    /**{ method
        @name stringOfBlanks
        @function return a string containing n blanks - hopfully this
             is efficient - especially for small n
        @param n - number blanks
        @return <Add Comment Here>
    }*/
    public static String stringOfBlanks(int n) {
        if(n <= 0) {
            return("");
        }
        switch(n) {
            case 1 :
            return(" ");
            case 2 :
            return("  ");
            case 3 :
            return("   ");
            case 4 :
            return("    ");
            case 5 :
            return("     ");
            case 6 :
            return("      ");
            case 7 :
            return("       ");
            case 8 :
            return("        ");
            case 9 :
            return("         ");
            case 10 :
            return("          ");
            case 11 :
            return("           ");
            case 12 :
            return("            ");
            default :
            return("            " + stringOfBlanks(n - 12));
        }
    }

    /**{ method
        @name displayString
        @function Break up NerdCapitalized words or _ separated words
             i.e. MyGoodName -> My Good Name
             JFK_Chair -> JFK Chair
        @param in - the string
        @return - display string
    }*/
    public static String displayString(String in) {
        int nchars = in.length();
        boolean PrevIsCapitalized = false;
        boolean PrevIsSpace = false;
        StringBuffer s = new StringBuffer(nchars);
        for(int i = 0; i < nchars; i ++) {
            char c = in.charAt(i);
            if(c == '_') {
                if(i != 0) {
                    // drop leading _
                    s.append(' ');
                    // convert _ to space
                    PrevIsSpace = true;
                }
            }
            else {
                // always capitalize the first char
                if(i == 0) {
                    s.append(Character.toUpperCase(c));
                }
                // any capitali
                else {
                    // for new upper - NerdCapitalization add a space
                    if(Character.isUpperCase(c) && ! PrevIsCapitalized && ! PrevIsSpace) {
                        s.append(' ');
                    }
                    s.append(c);
                }
                PrevIsSpace = false;
            }
            // remember if last is upper case so we do not
            // split acronyms such as LCJ
            PrevIsCapitalized = Character.isUpperCase(c);
        }
        return(s.toString());
    }
    /**{ method
        @name displayString
        @function Break up NerdCapitalized words or _ separated words
             i.e. MyGoodName -> MY_GOOD_NAME
             JFK_Chair -> JFK_CHAIR
        @param in - the string
        @return - display string
    }*/
    public static String fromDisplayString(String in) {
        int nchars = in.length();
        boolean PrevIsCapitalized = false;
        boolean PrevIsSpace = false;
        boolean IsCapitalized = false;
        StringBuffer s = new StringBuffer(nchars);
        for(int i = 0; i < nchars; i ++) {
            char c = in.charAt(i);
            IsCapitalized = Character.isUpperCase(c);

            if(IsCapitalized && !PrevIsCapitalized)
            {
                if(i != 0) {
                    // add leading _
                    s.append('_');
                    // remember to space
                    PrevIsSpace = true;
                }
                s.append(Character.toUpperCase(c));
           }
            else {
                // always capitalize the first char
                s.append(Character.toUpperCase(c));
                PrevIsSpace = false;
            }
            // remember if last is upper case so we do not
            // split acronyms such as LCJ
            PrevIsCapitalized = IsCapitalized;
        }
        return(s.toString());
    }

    /**{ method
        @name isEmptyString
        @function test if string null or 0 length
        @param s - the string
        @return - true if the string is null or 0 length
    }*/
    public static boolean isEmptyString(String s) {
        return(s == null || s.length() == 0);
    }

    /**{ method
        @name isAllUpperCase
        @function test if string is all upper case - empty returns false
        @param s - the string
        @return - true as above
    }*/
    public static boolean isAllUpperCase(String s) {
        if(isEmptyString(s)) {
            return(false);
        }
        for(int i = 0; i < s.length(); i ++) {
            if(Character.isLowerCase(s.charAt(i))) {
                return(false);
            }
        }
        return(true);
    }

    /**{ method
        @name isAllLowerCase
        @function test if string is all lower case - empty returns false
        @param s - the string
        @return - true as above
    }*/
    public static boolean isAllLowerCase(String s) {
        if(isEmptyString(s)) {
            return(false);
        }
        for(int i = 0; i < s.length(); i ++) {
            if(Character.isUpperCase(s.charAt(i))) {
                return(false);
            }
        }
        return(true);
    }

    /**{ method
        @name stripEnding
        @function drop an end fron a string i.e. foo.class -> foo
        @param in - the string to drop ending
        @param ending - ending to drop
        @return - string without the endinf
    }*/
    public static String stripEnding(String in,String ending)
    {
        if(in.endsWith(ending))
            return(in.substring(0,in.length() - ending.length()));
        return(in); // does not have ending
    }

    /**{ method
        @name isStringNumber
        @function test if string is a number can be -123.56
        @param s - the string
        @return - true as above
    }*/
    public static boolean isStringNumber(String s) {
        boolean hasDecimalPoint = false;
        if(isEmptyString(s)) {
            return(false);
        }
        for(int i = 0; i < s.length(); i ++) {
            char c = s.charAt(i);
            switch(c) {
                case '-' :
                // allow leading -
                if(i > 0) {
                    return(false);
                }
                break;
                case '.' :
                // allow 1 decimal point
                if(! hasDecimalPoint) {
                    hasDecimalPoint = true;
                }
                else {
                    return(false);
                }
                break;
                case '0' :
                // allow all digits
                case '1' :
                case '2' :
                case '3' :
                case '4' :
                case '5' :
                case '6' :
                case '7' :
                case '8' :
                case '9' :
                break;
                default :
                return(false);
            }
        }
        return(true);
        // nothing but digits and
    }

    /**{ method
        @name makeCharString
        @function Make a string with a single character
        @UnusedParam> - i the character
        @param i <Add Comment Here>
        @return - the string
    }*/
    public static String makeCharString(char i) {
        StringBuffer s = new StringBuffer(2);
        s.append((char) i);
        return(s.toString());
    }

    /**{ method
        @name isPrintingChar
        @function is a character printable
        @UnusedParam> - ch the character
        @param ch <Add Comment Here>
        @return - true if so
    }*/
    public static boolean isPrintingChar(char ch) {
        return(ch >= ' ' && ch <= 126);
    }

    /**{ method
        @name makeStringXYArray
        @function make a 2 dimensional array of rows and columns - all strings
             are created and set to "".
        @UnusedParam> - rows - number or rows
        @UnusedParam> - cols - number or columns
        @param rows <Add Comment Here>
        @param cols <Add Comment Here>
        @return - the array
    }*/
    public static String[][] makeStringXYArray(int rows,int cols) {
        String[][] ret = new String[rows][];
        for(int i = 0; i < rows; i ++) {
            ret[i] = new String[cols];
            for(int j = 0; j < cols; j ++) {
                ret[i][j] = "";
            }
        }
        return(ret);
    }

    /**{ method
        @name trimTrailingBlanks
        @function drop blank chars from the end but not the beginning of a string
             also filters non-printing characters
        @param s input string must be non-null
        @return a String holding the extracted characters - trimmed may be empty
    }*/
    public static String trimTrailingBlanks(String s) {
        int l = s.length() - 1;
        while(l >= 0) {
            if(s.charAt(l) != ' ') {
                break;
            }
            l --;
        }
        if(l < 0) {
            return("");
        }
        StringBuffer sb = new StringBuffer(l + 1);
        // now accumulate filtering out non-printing characters
        for(int i = 0; i <= l; i ++) {
            char c = s.charAt(i);
            if(c >= ' ' && c < 127) {
                sb.append(c);
            }
        }
        return(sb.toString());
    }
    /** method
   * drop white chars from the end but not the beginning of a string
   * also filters non-printing characters
   * @param s input string must be non-null
   *@return a String holding the extracted characters - trimmed may be empty
    */
    public static String trimWhiteSpace(String s) {
        int l = s.length();
        int start = 0;
        int end = l;
        for(; start < l; start++) {
            if(!Character.isWhitespace(s.charAt(start))) 
                break;
        }
        for(; end > start; end--) {
            if(!Character.isWhitespace(s.charAt(end - 1))) 
                break;
        }
        return(s.substring(start,end));
    }

    /**{ method
        @name parseInt
        @function just like Integer.parseInt but will quietly return 0
             is the string is empty
        @param s input string may be null
        @return a String holding the extracted characters - trimmed may be empty
    }*/
    public static int parseInt(String s) {
        if(s == null) {
            return(0);
        }
        if(s.length() == 0) {
            return(0);
        }
        return(Integer.parseInt(s));
    }

    /**{ method
        @name reverse
        @function - reverse the order of an array in place
        @param s the array
    }*/
    public static void reverse(Object[] s) {
        Object[] temp = new Object[s.length];
        System.arraycopy(s,0,temp,0,s.length);
        int k = s.length - 1;
        for(int i = 0; i < s.length; i ++)
            s[k --] = temp[i];
    }

    /**{ method
        @name reverse
        @function - reverse the order of an array in place
        @param s the array
    }*/
    public static void reverse(int[] s) {
        int[] temp = new int[s.length];
        System.arraycopy(s,0,temp,0,s.length);
        int k = s.length - 1;
        for(int i = 0; i < s.length; i ++)
            s[k --] = temp[i];
    }

    /**{ method
        @name reverse
        @function - reverse the order of an array in place
        @param s the array
    }*/
    public static void reverse(double[] s) {
        double[] temp = new double[s.length];
        System.arraycopy(s,0,temp,0,s.length);
        int k = s.length - 1;
        for(int i = 0; i < s.length; i ++)
            s[k --] = temp[i];
    }

    /**{ method
        @name reverse
        @function - reverse the order of an array in place
        @param s the array
    }*/
    public static void reverse(char[] s) {
        char[] temp = new char[s.length];
        System.arraycopy(s,0,temp,0,s.length);
        int k = s.length - 1;
        for(int i = 0; i < s.length; i ++)
            s[k --] = temp[i];
    }

    /**{ method
        @name reverse
        @function - reverse the order of an array in place
        @param s the array
    }*/
    public static void reverse(boolean[] s) {
        boolean[] temp = new boolean[s.length];
        System.arraycopy(s,0,temp,0,s.length);
        int k = s.length - 1;
        for(int i = 0; i < s.length; i ++)
            s[k --] = temp[i];
    }

    /**{ method
        @name cat
        @function - concatenate two arrays of objects
        @param arg1 first array
        @param arg2 first array
        @return - concatenated array
    }*/
    public static Object[] cat(Object[] arg1,Object[] arg2) {
        if(arg1 == null) {
            return(arg2);
        }
        if(arg2 == null) {
            return(arg1);
        }
        Object[] ret = new Object[arg1.length + arg2.length];
        System.arraycopy(arg1,0,ret,0,arg1.length);
        System.arraycopy(arg2,0,ret,arg1.length,arg2.length);
        return(ret);
    }

    /**{ method
        @name catStringArrays
        @function - concatenate two arrays of objects
        @param arg1 first array
        @param arg2 first array
        @return - concatenated array
    }*/
    public static String[] catStringArrays(String[] arg1,String[] arg2) {
        if(arg1 == null) {
            return(arg2);
        }
        if(arg2 == null) {
            return(arg1);
        }
        String[] ret = new String[arg1.length + arg2.length];
        System.arraycopy(arg1,0,ret,0,arg1.length);
        System.arraycopy(arg2,0,ret,arg1.length,arg2.length);
        return(ret);
    }


     /**{ method
        @name hidePropertyDescriptor
        @function - used to search a list of properties and set the property
           named in testName to hidden
        @param testName - Property name
        @param properties - list of all properties
    }*/
   public static void hidePropertyDescriptor(String testName,PropertyDescriptor[] properties) {
        for(int i = 0; i < properties.length; i++) {
            if(testName.equalsIgnoreCase(properties[i].getName())) {
                properties[i].setHidden(true);
                return;
            }
        }
    }
    
   /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static Object randomKey(Map in) {
        return(randomElement(in.keySet()));
    }
    
   /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static Object randomValue(Map in) {
        return(randomElement(in.values()));
    }
    
   /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static Object randomElement(Collection in) {
        int size = in.size();
        if(size == 0)
            return(null);
        int item = 0;
        if(size > 1)
            item = randomInt(size);
        List TheList = new ArrayList(in);
        Object ret = TheList.get(item);
        TheList = null;
        return(ret);
    }

     /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static Object randomElement(Object[] in) {
        if(in.length == 1)
            return(in[0]);
        return(in[randomInt(in.length)]);
    }

     /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static int randomIndex(Object[] in) {
        if(in.length == 1)
            return(0);
        return(randomInt(in.length));
    }

     /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static double randomElement(double[] in) {
        if(in.length == 1)
            return(in[0]);
        return(in[randomInt(in.length)]);
    }

     /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static int randomElement(int[] in) {
        if(in.length == 1)
            return(in[0]);
        return(in[randomInt(in.length)]);
    }

     /**{ method
        @name randomElement
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static int randomInt(int in) {
        if(in <= 1 ) {
            if(in == 1)
                return(0);
            throw new IllegalArgumentException("randomInt must take a number > 1");
        }
        int test = gRandomizer.nextInt();
        if(test < 0)
            test = -test;
        return(test % in);
    }

     /**{ method
        @name randomString
        @function - choose an element of an array at random
        @param in - choices - non-null non-empty
        @return - one choice
    }*/
    public static String randomString() {
        return(randomString(8));
    }
     /**{ method
        @name randomString
        @function - create a randomString of letters of length n
        @param in - positive int
        @return - the string length 1..256
    }*/
    public static String randomString(int in) {
        in = Math.max(1,Math.min(in,256));
        StringBuffer sb = new StringBuffer();
        for(int i = 0; i < in; i++) {
            char c = (char)('A' + randomInt(26));
            sb.append(c);
        }
        return(sb.toString());
    }
    /**
        like math.random but will not be the same on multiple runs
        @return - random number 0..1
    */
    public static double random()
    {
        return(gRandomizer.nextDouble());
    }

    /**
         randomGaussian returna variable with a guassian distribution
        @param mean - distribution mean
        @param sd - distribution sd
        @return - one variable
    */
    public static double randomGaussian(double mean,double sd)
    {
        double norm = randomNormalGaussian();
        return((norm * sd) + mean);
    }

    /**
    *  randomGaussian returna variable with a guassian distribution
    * Test program to generate a vector of Gaussian deviates using
    *  the Box-Muller method and the system-supplied uniform RNG.
    *  K.A.Hawick, 15 July 1994.
    *
    *  H W Yau.  23rd of August, 1996.
    *  Northeast Parallel Architectures Center.
    *  Syracuse University.
    *  @return - one variable
    */
    public static double randomNormalGaussian()
    {
        double r = 0;
        double f = 0;
        double v1 = 0;
        double v2 = 0;
        do {
            v1 = gRandomizer.nextDouble();
            v2 = gRandomizer.nextDouble();
            v1 = (2.0 * v1) -1.0;
            v2 = (2.0 * v2) -1.0;
            r = ((v1 * v1) + (v2 * v2));
            f = Math.sqrt( (-2.0 * Math.log(r)) / r);
        }
        while(r >= 1.0 || r <= 0);

        double ret = v1 * f;
        if(Double.isNaN(ret))
            Assertion.doNada();

        return(ret);
    }

     /**{ method
        @name getPopertiesList
        @function - A list is stored as Properties called
	   <ListName>.0
	   <ListName>.1
	   <ListName>.2 ...
	   This takes listName and returns an array os Strings with
	   0 or more elements
        @param prop - the properties
	@param name - base name
        @return - the list
    }*/
    public static String[] getPropertiesList(Properties prop, String Name) {
        Vector holder = new Vector();
	    int index = 0;
	    String testname = Name + "." + index++;
	    String test = (String)prop.get(testname);
	    while(test != null) {
	        holder.addElement(test);
	        test = (String)prop.get(Name + "." + index++);
	        }
	    String[] ret = new String[holder.size()];
	    holder.copyInto(ret);
	    return(ret);
    }


     /**{ method
        @name setPopertiesList
        @function - A list is stored as Properties called
	   <ListName>.0
	   <ListName>.1
	   <ListName>.2 ...
	   This takes listName and returns an array os Strings with
	   0 or more elements
        @param prop - the properties
	@param name - base name
	@param data - array of 0 or more strings
    }*/
    public static void setPropertiesList(Properties prop, String Name,String[] data) {
	    int index = 0;
	    // Wipe out old values
	    String TestName = Name + "." + index++;
	    String test = (String)prop.get(TestName);
	    while(test != null) {
	        prop.remove(TestName);
	        TestName = Name + "." + index++;
	        test = (String)prop.get(TestName);
	    }
	    for(int i = 0; i < data.length; i++)
	        prop.put(Name + "." + i,data[i]);
    }

     /**{ method
        @name makeClassList
        @function - convert a list of Objects to a list of
        clas names
	    @param in -the list - may be null
	    @return = non-null list of class names
    }*/
    public static String[] makeClassList(Object[] in)
    {
        if(in == null)
            return(new String[0]);
        String[] ret = new String[in.length];
	    for(int i = 0; i < in.length; i++)
	        ret[i] = in[i].getClass().getName();
	    return(ret);
    }

     /**{ method
        @name intString
        @function - return a string representing an int which is
        RequiredPlaces long (or longer) padding with 0 s
        Yes format can to this but this function shuld be 1.02
        conpatable.
        NOTE - best used for positive data
        @param data - the int
        @param RequiredPlaces - the required width
	    @param return  - resultant string
    }*/
    public static String intString(int data,int RequiredPlaces)
    {
        String out = Integer.toString(data);
        while(out.length() <  RequiredPlaces)
            out = "0" + out;
        return(out);
    }

     /**{ method
        @name listMember
        @function - test of the object test is a member of the list data
        @param test - Object to test
        @param data - Array of possibilities
	    @param return  - true if included
    }*/
    public static boolean listMember(Object test,Object[] data)
    {
        for(int i = 0; i < data.length; i++) {
            if(test == data[i])
                return(true); // member
        }
        return(false); // not found
    }

     /**{ method
        @name capitalize
        @function - test of the object test is a member of the list data
        @param in - string to capitalize
 	    @param return  - capil\tailzed string
    }*/
    public static String capitalize(String in)
    {
        if(in.length() == 0)
            return("");
        StringBuffer s = new StringBuffer(in.length());
        int i = 0;
        char c = in.charAt(i++);
        while(c <= ' ') {
            c = in.charAt(i++);
            if(i >= in.length())
                return("");
        }

        s.append(Character.toUpperCase(c));
        boolean atSpace = false;
        for(; i <  in.length(); i++) {
            c = Character.toLowerCase(in.charAt(i));
            if(c == ' ') {
                atSpace = true;
            }
            else {
                if(atSpace) {
                    if(c > ' ') {
                        c = Character.toUpperCase(c);
                        atSpace = false;
                    }
                }
            }
            s.append(c);
        }
        return(s.toString());
    }
 
     /**{ method
        @name nerdCapsToWords
        @function - convert FieFieFoe to Fie Fie Foe
        NOTE Seqiences i.e. XML will be concatinated
        @param in - string to capitalize shouldnot have spaces
 	    @param return  - broken string
    }*/
    public static String nerdCapsToWords(String in)
    {
        in = in.trim();
        if(in.length() == 0)
            return("");
        StringBuffer s = new StringBuffer(in.length() + 10);
        int i = 0;
        char c = in.charAt(i++);
        boolean LastWasCap = true;
        s.append(Character.toUpperCase(c));
        for(; i <  in.length(); i++) {
            c = in.charAt(i);
            if(Character.isUpperCase(c)) {
                if(!LastWasCap)
                    s.append(' '); // add space
            }
            else {
                LastWasCap = false;
            }
            s.append(c);
        }
        return(s.toString());
    }
     /**{ method
        @name nerdCapsToWords
        @function - convert FieFieFoe to Fie Fie Foe
        NOTE Seqiences i.e. XML will be concatinated
        @param in - string to capitalize shouldnot have spaces
 	    @param return  - broken string
    }*/
    public static String wordsToNerdCaps(String in)
    {
        in = in.trim();
        StringBuffer s = new StringBuffer(in.length() + 10);
        boolean LastWasSpace = false;
        for(int i = 0; i <  in.length(); i++) {
            char c = in.charAt(i);
            if(Character.isSpaceChar(c)) {
                LastWasSpace = true;
            }
            else {
                if(LastWasSpace) {
                    s.append(Character.toUpperCase(c));
                }
                else {
                     s.append(c);
                }
            }
        }
        return(s.toString());
    }

   /**
         return a string with topic names separated by a delimiter -
         i.e. "," or "\t"
         @param items of options if null an empty string is returned
         @param delimiter non-null separater string
         @return the string non-null
    */
    public static String buildDelimitedString(String[] items,String delimiter)
    {
        StringBuffer ret = new StringBuffer();
        if(items != null) {
            for(int i = 0; i < items.length; i++) {
                if(i > 0)
                    ret.append(delimiter);
                ret.append(items[i]);
            }
        }
        return(ret.toString());
    }
    
    /**
    *     build indent String
    *     @param level non-negative every level indents 4 spaces
    *     @return string with indented CR
    */
    public static String indentString(int indent) 
    {
        if(gIndentStrings == null) {
            buildIndentStrings();
        }
        if(indent < gIndentStrings.length)
            return(gIndentStrings[indent]);
        // too long so build
        StringBuffer sb = new StringBuffer(4 * indent);
        for(int i = 0; i < indent; i++)
            sb.append("    ");
        return(sb.toString());
        
    }
    /**
    *     build indent String
    *     @param level non-negative every level indents 4 spaces
    *     @return string with indented CR
    */
    protected static synchronized void buildIndentStrings() 
    {
        if(gIndentStrings != null) {
            return;
        }
        String[] TheStrings = new String[16];
        String IndentString = "";
        for(int i = 0; i < TheStrings.length; i++) {
            TheStrings[i] = IndentString;
            IndentString += "    ";
        }
        gIndentStrings = TheStrings;
    }
    
    /**
    *     add indent to a String
    *     @param out non-null PrintStream
    *     @param level non-negative every level indents 4 spaces
    *     @return string with indented CR
    */
    public static String indentTransform(String in,int indent) 
    {
        StringBuffer sb = new StringBuffer(in.length() + 64 * indent);
        String IndentString = indentString(indent);
        for(int i = 0; i < in.length(); i++) {
            char c = in.charAt(i);
            switch(c) {
                case '\n' :
                    sb.append(c);
                    if(i < (in.length() - 1))
                        sb.append(IndentString);
                    break;
                case '\t' :
                    sb.append("    ");
                    break;
                default :
                    if(c >= ' ')
                        sb.append(c);
            }
        }
        return(sb.toString());
    }

    /**
    *     add indent to a stream
    *     @param out non-null PrintStream
    *     @param level non-negative every level indents 4 spaces
    */
    public static void indent(PrintStream out,int level)
    {
        out.print(indentString(level));
    }

    /**
    *     add indent to a stream
    *     @param out non-null PrintWriter
    *     @param level non-negative every level indents 4 spaces
    */
    public static void indent(PrintWriter out,int level)
    {
        out.print(indentString(level));
    }
    /**
         Sleep without worrying about try .. catch - one line sleep
         @param time - sleep time
    */
    public void doSleep(int time)
    {
        try {
            Thread.sleep(time);
        }
        catch(InterruptedException ex) {
            throw new ThreadInterruptedException();
        }
    }
    
    
      /**
      *   Sleep forever or until interrupted
      */
      public static void sleepForever()
      {
        while(true) {
            try {
                Thread.sleep(2000);
            }
            catch(InterruptedException ex) {
                 throw new ThreadInterruptedException();
            }
        }
      }



	/**
	 * helper function to be used in implementing Comparable
	 * @param t1 number to compare
	 * @param t2 number to compare
	 * @return 0 if t1 == t2 1 if t1 > t2 else -1
	 */
	public static int makeComparison(int t1,int t2)
	{
	    if(t1 == t2)
	        return(0);
	    if(t1 < t2)
	        return(1);
	    else
	        return(-1);
	}


	/**
	 * helper function to be used in implementing Comparable
	 * @param t1 number to compare
	 * @param t2 number to compare
	 * @return 0 if t1 == t2 1 if t1 > t2 else -1
	 */
	public static int makeComparison(double t1,double t2)
	{
	    if(t1 == t2)
	        return(0);
	    if(t1 < t2)
	        return(1);
	    else
	        return(-1);
	}

	/**
	 * helper function to be used in implementing Comparable
	 * @param t1 number to compare
	 * @param t2 number to compare
	 * @return 0 if t1 == t2 1 if t1 > t2 else -1
	 */
	public static int makeComparison(long t1,long t2)
	{
	    if(t1 == t2)
	        return(0);
	    if(t1 < t2)
	        return(1);
	    else
	        return(-1);
	}

	/**
	 * Utility method to test equality on two possibly null objects
	 * @param in non-null string
	 * @param token token to break - if null use \n
	 * @return non-null
	 */
	public static String[] parseString(String in,String token)
	{
		if(token == null)
		    token = "\n";
		StringTokenizer st = new StringTokenizer(in,token);
		String[] ret = new String[st.countTokens()];
		int i = 0;
		while(st.hasMoreTokens())
		    ret[i++] = st.nextToken();
		return(ret);
	}


	/**
	 * Utility method to test equality on two possibly null objects
	 * @param t1 - possibly null test object
	 * @param t2 - possibly null test object
	 * @return true if t1 == t2 == null || t1.equals(t2)
	 */
	public static boolean objectEqual(Object t1, Object t2) {
		if (t1 == null) {
			return (t2 == null);
		}

		if (t2 == null) {
			return (false);
		}

		return (t1.equals(t2));
	}

	/**
	 * Utility method to return the hashCode of a possibly null object
	 * @param t1 - possibly null test object
	 * @return hash code or 0 if t1 == null
	 */
	public static int objectHash(Object t1) {
		if (t1 == null) {
			return (0);
		}

		return (t1.hashCode());
	}

    /**
    * round in up to the next even multiple of del
    * @param in initial value
    * @param del interval
    * @return next multiple of del >= in
    */
    public static double roundUpTo(double in, double del)
    {
        int n = (int)(in / del);
        double ret = n * del;
        if(ret == in)
            return(ret);
        return(ret + del);
    }
    /**
    * round in up to the next even multiple of del
    * @param in initial value
    * @param del interval
    * @return next multiple of del >= in
    */
    public static int roundUpTo(int in, int del)
    {
        int n = (int)(in / del);
        int ret = n * del;
        if(ret == in)
            return(ret);
        return(ret + del);
    }
    
    /**
    * remove leading and trailing quotes
    * @param in npn-null string
    * @return new string
    */
    public static String dropQuotes(String in)
    {
        String ret = in.trim();
        if(ret.length() < 2) 
            return(ret);
        char start = ret.charAt(0);
        char end = ret.charAt(ret.length() - 1);
        if(start != end)
            return(ret); 
        if(start == '\'' || start == '\"')
            return(ret.substring(1,ret.length() - 1));
        else
            return(ret);
    }

     /**
    * Make a string into a a multiline]
    * quoted string
    * @return new string
    */

    public static String multilineQuotedString(String in,int indent)
    {
        if(in.length() == 0)
            return("\"\"");

        StringTokenizer st = new StringTokenizer(in,"\n");
        boolean NotFirstLine = false;
        int tokenCount = st.countTokens();
        if(tokenCount == 1)
            return("\"" + in + "\"");
        int index = 1;
        StringBuffer holder = new StringBuffer(in.length() + 10);
        while(st.hasMoreTokens()) {
            if(NotFirstLine) {
                holder.append("\n");
                for(int i = 0; i < indent; i++)
                    holder.append("    ");
            }
            else {
                NotFirstLine = true;
            }
            holder.append("\"");
            holder.append(st.nextToken());
            holder.append("\\n\"");
            if(index++ < tokenCount)
                holder.append(" + ");
        }
        return(holder.toString());
    }

    /**
    * print the lines in CodeString with
    * indentation given of indent
    * @out non-null writer
    * @CodeString  non-null possibly multi=line string
    * @indent non-negative indent value
    */
    public static void printIndentedCode(PrintWriter out,String CodeString,int indent)
    {
        StringTokenizer st = new StringTokenizer(CodeString,"\n");
        while(st.hasMoreTokens()) {
            Util.indent(out,indent);
            out.println(st.nextToken());
        }
    }

    /**
    * Makes an iterator over an array
    * @param in - non-null array of Objects
    * @return - non-null iterator over the array elements
    */
    public static Iterator arrayIterator(Object[] array)
    {
        List TheList = Arrays.asList(array);
        return(TheList.iterator());
    }

    /**
    * makes a copy of any array returning the same type
    * @param in - non-null array of unknown type
    * @return - non-null copy array of input type
    */
    public static Object duplicateArray(Object in)
    {
        Class inClass = in.getClass();
        if(!inClass.isArray())
            throw new IllegalArgumentException("must pass array");
        int InSize = Array.getLength(in); // size of input array
        Object ret = Array.newInstance(inClass.getComponentType(),InSize);
        System.arraycopy(in, 0, ret, 0, InSize);
     //   validateArrayNonNull(ret);
        return(ret);
    }


    /**
    * adds one object to an array - Yes this is inefficient
    * but usually arrays are accessed much more frequently than they grow
    * @param in - non-null array of unknown type
    * @param addTo - non-null object to add to the array
    * @return - non-null array of input type
    */
    public static Object removeFromArray(Object in,Object removeFrom) {
        Class inClass = in.getClass();
        if(!inClass.isArray())
            throw new IllegalArgumentException("must pass array");
        int InSize = Array.getLength(in); // size of input array
        int index = -1;
        for(int i = 0; i < InSize; i++) {
            if(removeFrom.equals(Array.get(in,i))) {
                index = i;
                break;
            }
        }
        if(index == -1)
            return(in); // not found

        Object ret = Array.newInstance(inClass.getComponentType(),InSize - 1);
        if(index == 0) {
            System.arraycopy(in, 1, ret, 0, InSize - 1);
      //      validateArrayNonNull(ret);
            return(ret);
        }
        if(index == InSize - 1) {
            System.arraycopy(in, 0, ret, 0, InSize - 1);
        //    validateArrayNonNull(ret);
            return(ret);
        }
        System.arraycopy(in, 0, ret, 0, index);
        System.arraycopy(in, index + 1, ret, index, InSize - 1 - index);
     //   validateArrayNonNull(ret);
        return(ret);

    }

    /**
    * return true if all elements of the
    * object as an array
    * are not null
    * @return as above
    */
    public static boolean validateArrayNonNull(Object in)
    {
       for(int i = 0; i < Array.getLength(in); i++) {
            if(Array.get(in,i) == null) {
                return(false);
            }
        }
        return(true);
    }

    /**
    * adds one object to an array - Yes this is inefficient
    * but usually arrays are accessed much more frequently than they grow
    * @param in - non-null array of unknown type
    * @param addTo - non-null object to add to the array
    * @return - non-null array of input type
    */
    public static Object[] addToArray(Object[] in,Object addTo) {
        Class inClass = in.getClass();
        if(!inClass.isArray())
            throw new IllegalArgumentException("must pass array");
        int InSize = Array.getLength(in); // size of input array
        Object[] ret = (Object[])Array.newInstance(inClass.getComponentType(),InSize + 1);
        System.arraycopy(in, 0, ret, 0, InSize);
        Array.set(ret,InSize,addTo);
   //     validateArrayNonNull(ret);
        return(ret);
    }
    /**
    * create e new array holding a segment of an original array 
    * which is the same type as the original
    * @param in - non-null array of unknown type
    * @param begin start of copy segment
    * @param end end of segment + 1
    * @return - non-null array of input type
    */
    public static Object[] copyArraySegment(Object[] in,int begin,int end)
    {
        if(end == 0)
            return(in);
        Class inClass = in.getClass();
        if(!inClass.isArray())
            throw new IllegalArgumentException("must pass array");
        int InSize = Array.getLength(in); // size of input array
        int RealEnd = Math.min(end,InSize);
        if(begin < 0)
            throw new IllegalArgumentException("begin must be greater than 0");
        int TargetSize = RealEnd - begin;
        if(begin >= RealEnd)
            throw new IllegalArgumentException("begin less than end"); 
           
        Object[] ret = (Object[])Array.newInstance(inClass.getComponentType(),TargetSize);
        System.arraycopy(in, begin, ret, 0, TargetSize);
        return(ret);
    }
    /**
    * doubles the sice if an input array
    * @param in - non-null array of unknown type
    * @return - non-null array of twice size of in
    */
    public static Object doubleArray(Object in) {
        Class inClass = in.getClass();
        if(!inClass.isArray())
            throw new IllegalArgumentException("must pass array");
        int InSize = Array.getLength(in); // size of input array
        Object ret = Array.newInstance(inClass.getComponentType(),
                        2 * InSize);
        System.arraycopy(in, 0, ret, 0, InSize);
        return(ret);
    }

    /**
    * convert a collection into an array of objects
    * @param in - non-null collection
    * @param type - non-null class must match all members of the collection
    * @return non-null array of type type
    */
    public static Object[] collectionToArray(Collection in,
			    Class type)
    {
	    Object[] ret = (Object[] )Array.newInstance(
				    type,in.size());
        in.toArray(ret);
        return(ret);
    }
    protected PrintWriter gSystemWriter;

    /**
    * convert a properties value into an int
    * @param p non-null properties
    * @param key non-null id of an existing key
    * @return int value - -1 if the key does not exist
    * @throws NumberFormatException if the value is not a number
    */
    public static int getPropertyInt(Properties p, String key)
    {
        Object test = p.get(key);
        if(test == null)
            return(-1);
        return(Integer.parseInt(test.toString()));
    }
    /**
    * convert a properties value into an int
    * @param p non-null properties
    * @param key non-null id of an existing key
    * @return double value - MIN_VALUE if the key does not exist
    * @throws NumberFormatException if the value is not a number
    */
    public static double getPropertyDouble(Properties p, String key)
    {
        Object test = p.get(key);
        if(test == null)
            return(Double.MIN_VALUE);
        return(Double.parseDouble(test.toString()));
    }

    /**
    * code snippet to load Properties from a resource file
    * @param in non-null object
    * @param name non-null existing resurce name
    * @return non-null properties
    * @throws OnviaException on error
    */
    public static Properties loadResourceProperties(Object in,String name)
    {
        return(loadResourceProperties(in.getClass(),name));
    }

    /**
    * code snippet to load Properties from a resource file
    * @param in non-null class
    * @param name non-null existing resurce name
    * @return non-null properties
    * @throws OnviaException on error
    */
    public static Properties loadResourceProperties(Class in,String name)
    {
        Properties ret = new Properties();
        try {
            InputStream inStream = in.getResourceAsStream(name);
            ret.load(inStream);
            return(ret);
        }
        catch(IOException ex) {
            Assertion.fatalException(ex);
            return(null);
        }
    }


    /**
    * get a printwriter to system.out
    * @return non-null writer to system.out
    */
    public PrintWriter systemWriter()
    {
        if(gSystemWriter == null)
            buildSystemWriter();
        return(gSystemWriter);
    }

    public synchronized void buildSystemWriter()
    {
        if(gSystemWriter != null)
            return;
        PrintWriter TheWriter = new PrintWriter(System.out);
        gSystemWriter = TheWriter;
    }

    /**
    * convert a double into a String with a given precision
    * default double formatting is not very pretty
    * @param in non-null Double to convert
    * @param int positive precision
    * @return non-null formatted string
    */
    public static String formatDouble(Double in,int precision)
    {
        return(formatDouble(in.doubleValue(),precision));
    }

    /**
    * convert a double into a String with a given precision
    * default double formatting is not very pretty
    * @param in double to convert
    * @param int positive precision
    * @return non-null formatted string
    */
    public static String formatDouble(double in,int precision)
    {
        NumberFormat nf = NumberFormat.getInstance();
        nf.setMaximumFractionDigits(precision);
        return(nf.format(in));
    }

    /**
    * convert a double into a String with a given precision
    * default double formatting is not very pretty
    * This version defaults precision to 2
    * @param in non-null Double to convert
    * @param int positive precision
    * @return non-null formatted string
    */
    public static String formatDouble(Double in)
    {
        return(formatDouble(in.doubleValue()));
    }

    /**
    * convert a double into a String with a given precision
    * default double formatting is not very pretty
    * This version defaults precision to 2
    * @param in double to convert
    * @return non-null formatted string
    */
    public static String formatDouble(double in)
    {
        return(formatDouble(in,2));
    }

    /**
    * return a string a time interval in millisec
    * @param in positive interval millisec
    * @returb non-null string i.e. 3:40
    */
    public static String formatTimeInterval(long in)
    {
        int del = (int)(in + 500) / 1000; // drop millisec
        int seconds = del % 60;
        del /= 60;
        int minute = del % 60;
        del /= 60;
        int hour = del % 24;
        del /= 24;
        int days = del;
        StringBuffer holder = new StringBuffer();

        if(days != 0) {
            holder.append(Integer.toString(days));
            holder.append(":");
            holder.append(d02(hour));
            holder.append(":");
            holder.append(d02(minute));
            holder.append(":");
            holder.append(d02(seconds));
        }
        else {
            if(hour != 0) {
                holder.append(Integer.toString(hour));
                holder.append(":");
                holder.append(d02(minute));
                holder.append(":");
                holder.append(d02(seconds));
            }
            else {
                holder.append(d02(minute));
                holder.append(":");
                holder.append(d02(seconds));
            }
        }
       return(holder.toString());
     }

    /**
    * return a string represtneing the hours and minutes
    * of a Data
    * @param in non-null data
    * @returb non-null string i.e. 3:40
    */
    public String formatMinutes(java.util.Date in)
    {
        GregorianCalendar TheCalendar = new GregorianCalendar();
        TheCalendar.setTime(in);
        int hour = TheCalendar.get(Calendar.HOUR);
        int minute  = TheCalendar.get(Calendar.MINUTE);
        StringBuffer holder = new StringBuffer();
        holder.append(Integer.toString(hour));
        holder.append(":");
        holder.append(d02(minute));
        return(holder.toString());
    }

    /**
    * convert a 2 digit int to 00 , 02 or 12
    * in integer between 0 and 99
    * @return non-null string
    */
    protected static String d02(int in)
    {
        if(in == 0) {
            return("00");
        }
        else {
            if(in < 10) {
                return("0" + in);
            }
            return(Integer.toString(in));
        }
    }
    /**
    * return a string represtneing the hours and minutes and seconds
    * of a Data
    * @param in non-null data
    * @returb non-null string i.e. 3:40
    */
    public static String formatSeconds(java.util.Date in)
    {
        GregorianCalendar TheCalendar = new GregorianCalendar();
        TheCalendar.setTime(in);
        int hour = TheCalendar.get(Calendar.HOUR);
        int minute  = TheCalendar.get(Calendar.MINUTE);
        int seconds = TheCalendar.get(Calendar.SECOND);
        StringBuffer holder = new StringBuffer();
        holder.append(Integer.toString(hour));
        holder.append(":");
        holder.append(d02(minute));
        holder.append(":");
        holder.append(d02(seconds));
       return(holder.toString());
    }

    /**
    * prints all strings one per line
    * @param out non-null writer
    * @param items non-null array of strings
    */
    public static void printItems(PrintWriter out,String[] items )
    {
        for(int i = 0; i < items.length; i++)
            out.println(items[i]);
    }
    /**
    * prints all strings one per line
    * @param out non-null writer
    * @param items non-null array of strings
    */
    public static void printItems(PrintStream out,String[] items )
    {
        for(int i = 0; i < items.length; i++)
            out.println(items[i]);
    }
    /**
    * return true if ignoring white space and non-printing characters
    * the two arguments are equivalent strings
    * @param s1 non-null string
    * @param s2 non-null string
    * @return as above
    */
    public static boolean equivalentForHtml(String s1,String s2) {
        return(stringWithoutHTMLComments(s1).equals(stringWithoutHTMLComments(s2)));
    }

    /**
    * return true if ignoring white space and non-printing characters
    * the two arguments are equivalent strings
    * @param s1 non-null string
    * @param s2 non-null string
    * @return as above
    */
    public static boolean equivalentIgnoreWhite(String s1,String s2) {
        return(stringWithoutWhite(s1).equals(stringWithoutWhite(s2)));
    }
    
    /**
    * Make a string dropping all white and non-printing characters
    * @param s1 non-null input string
    * @return non-null string as above
    */
    public static String stringWithoutWhite(String s1) {
        StringBuffer sb = new StringBuffer();
        for(int i = 0; i < s1.length(); i++) {
           char c = s1.charAt(i);
           if(c > ' ') {
               sb.append(c);
           }
        }
        return(sb.toString());
    }
    /**
    * Make a string dropping all white and non-printing characters and HTML Comments
    * @param s1 non-null input string
    * @return non-null string as above
    */
    public static String stringWithoutHTMLComments(String s1) {
        int index = s1.indexOf("<!--");
        String temp = s1;
        while(index > -1) {
            if(index > 0)
                s1 = s1.substring(0,index);
            else 
                s1 = "";
            int endIndex = temp.indexOf("-->",index);
            if(endIndex > index) {
                if(endIndex + 3 < temp.length())
                    s1 += temp.substring(endIndex + 3);
            }
            temp = s1;
            index = s1.indexOf("<!--");
        }
        return(stringWithoutWhite(s1));
    }
    
    /**
    * return true if the character is not part of a word
    * used primarily in replaceWordInString
    * @param c test character
    * @return as above
    */
    public static boolean isWordTerminator(char c)
    {
        if(Character.isWhitespace(c))
            return(true);
        if(Character.isJavaLetterOrDigit(c))
            return(false);
        return(true);
    }    
    
    /**
    * Replace All  instances of ReplaceText starting at start
    * this differs from replaceInString in that the text must be followed by a wordTerminator
    * @param s1 non-null input string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    */
    public static String replaceWordInString(String s1,String ReplaceText,String NewText) {
        return(replaceWordInString(s1,ReplaceText, NewText,0));
    }


    /**
    * Replace All  instances of ReplaceText starting at start
    * this differs from replaceInString in that the text must be followed by a wordTerminator
    * @param s1 non-null input string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    */
    public static String replaceWordInString(String s1,String ReplaceText,String NewText,int start) {
        int index = s1.indexOf(ReplaceText, start);
        int ReplaceLength = ReplaceText.length();
        if(index == -1)
            return(s1.substring(start));
        StringBuffer sb = new StringBuffer(s1.length());
        if(s1.length() == index + ReplaceLength || isWordTerminator(s1.charAt(index + ReplaceLength))) { // match is word
            sb.append(s1.substring(start,index));
            sb.append(NewText);
            sb.append(replaceWordInString(s1,ReplaceText,NewText,index + ReplaceLength));
        }
        else { // match is not word
            sb.append(s1.substring(start,index  + ReplaceLength));
            sb.append(replaceWordInString(s1,ReplaceText,NewText,index + ReplaceLength));
       }
        return(sb.toString());
    }


    /**
    * Replace All  instances of ReplaceText starting at start
    * @param s1 non-null input string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    */
    public static String replaceInString(String s1,String ReplaceText,String NewText) {
        return(replaceInString(s1,ReplaceText, NewText,0));
    }

    /**
    * Replace All  instances of ReplaceText starting at start
    * @param s1 non-null input string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    */
    public static String replaceInString(String s1,String ReplaceText,String NewText,int start) {
        int index = s1.indexOf(ReplaceText, start);
        if(index == -1)
            return(s1.substring(start));
        StringBuffer sb = new StringBuffer(s1.length());
        sb.append(s1.substring(start,index));
        sb.append(NewText);
        sb.append(replaceInString(s1,ReplaceText,NewText,index + ReplaceText.length()));
        return(sb.toString());
    }

    /**
    * Replace first instance of ReplaceText starting at start - 
    * @param s1 non-null input string must contain the replace string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    * @throws IllegalArgumentException if the substring is not found
    */
    public static String replaceOnceInStringRequired(String s1,String ReplaceText,String NewText) {
        int index = s1.indexOf(ReplaceText);
        if(index == -1) // not found
            throw new IllegalArgumentException("String '" + ReplaceText +
                "' not found in string '" + s1 + "'");
        return(replaceOnceInString(s1,ReplaceText,NewText));
    }

    /**
    * Replace first instance of ReplaceText starting at start - if replacetext not found 
    * return the original string
    * @param s1 non-null input string may contain the replace string
    * @param ReplaceText non-null TextToReplace
    * @param NewText non-null TextTo Insert
    * @return non-null string as above
    */
    public static String replaceOnceInString(String s1,String ReplaceText,String NewText) {
        int index = s1.indexOf(ReplaceText);
        if(index == -1) // not found
            return(s1);
        StringBuffer sb = new StringBuffer(s1.length());
        if(index > 0)
            sb.append(s1.substring(0,index));
        sb.append(NewText);
        if(index < s1.length() - ReplaceText.length())
            sb.append(s1.substring(index + ReplaceText.length()));
        return(sb.toString());
    }

    /**
    * return the number of times the substring occurs in the text - 
    * where there are overlaps the count tells how many times the substring could be replaced
    * @param s1 non-null input string to test
    * @param SubString non-null Substring to cound
    * @return non-negative number of occurrances of the substring
    */
    public static int countSubstringOccurrances(String s1,String SubString) {
        int count = 0;
        int index = s1.indexOf(SubString);
        int sublength = SubString.length();
        int start = 0;
        while(index > -1) {
            count++;
            start = index + sublength;
            if(start > s1.length() - sublength)
                return(count);
            index = s1.indexOf(SubString,start);
        }
        return(count);
    }
    
    /**
    * return the keys in a Map as a StringArray
    * @in non-null map
    * @return non-null array of keys as strings
    */
    public static String[] getKeyStrings(Map in) 
    {
        String[] ret = new String[in.size()];
        SortedSet keys = new TreeSet(in.keySet());
        Iterator it = keys.iterator();
        int count = 0;
        while(it.hasNext())
            ret[count++] = it.next().toString();
        return(ret);
    }
    
    /**
    * return the values in an array of NameValues
    * @in non-null map
    * @return non-null array of NameValues where name is the key valye the value
    */
    public static NameValue[] mapToNameValues(Map in) 
    {
        String[] keys = getKeyStrings(in);
        NameValue[] ret = new NameValue[keys.length];
        for(int i= 0; i < keys.length; i++) {
            ret[i] = new NameValue(keys[i],in.get(keys[i]));
        }
        return(ret);    
    }
    
   /**
    * convert array of NameValues to a Map
    * @in non-null non-null array of NameValues
    * @return non-null Map with names as keys values as values
    */
    public Map nameValuesToMap(NameValue[] in)
    {
        Map ret = new HashMap();
        for(int i= 0; i < in.length; i++) {
            ret.put(in[i].m_Name,in[i].m_Value);
        }
        return(ret);    
    }
    
    /**
    * return the values in a Map as a StringArray
    * @in non-null map
    * @return non-null array of keys as strings
    */
    public static String[] getValueStrings(Map in) 
    {
        String[] ret = new String[in.size()];
        SortedSet keys = new TreeSet(in.keySet());
        Iterator it = keys.iterator();
        int count = 0;
        while(it.hasNext())
            ret[count++] = it.next().toString();
        return(ret);
    }
    
    /**
    * builde a well formatted multi-line list of items wrapping every 
    * 8 items
    * @param items - non-null array of items to add
    * @parem indent non-negative indent level
    * @return non-null string with all iwtms
    */
    public static String buildListString(Object[] items,int indent) 
    {
        return(buildListString(items,indent,8));
    }
    
    
    /**
    * builde a well formatted multi-line list of items wrapping every 
    * 8 items
    * @param items - non-null array of items to add
    * @parem indent non-negative indent level
    * @parem itemsPerLine positive number of items per line
    * @return non-null string with all iwtms
    */
    public static String buildListString(Object[] items,int indent,int itemsPerLine) 
    {
        StringBuffer Holder = new StringBuffer();
        for(int i = 0; i < items.length; i++) {
            if((i % itemsPerLine) == 0) {
                Holder.append("\n");
                Holder.append(indentString(indent));
            }
            else {
                Holder.append(", ");
            }
            Holder.append(items[i].toString());
        }
            
        return(Holder.toString());
    }
    
    /**
    * return the first value in a map - this is most useful where only one item 
    * exists in th emap
    * @param in non-null Map with at least one value
    * @return non-null object
    * @throws IllegalArgumentException of the map is empty
    */
    public static Object firstValue(Map in) 
    {
        return(firstElement(in.values()));
    }
    
    /**
    * return the first element in a collection - this is most useful where only one item 
    * exists in the collection
    * @param in non-null collection with at least one value
    * @return non-null object
    * @throws IllegalArgumentException of the collection is empty
    */
    public static Object firstElement(Collection in) 
    {
        if(in.size() == 0)
            throw new IllegalArgumentException("firstElement requies at least one entry");
        return(in.iterator().next());
    }
    /**
    * convert an array of objects to a List
    * @param non-null array
    * @return non-null list
    */
    public static List makeList(Object[] items)
    {
       return(Arrays.asList(items));
    }
    /**
    * convert an array of objects to a List
    * @param non-null array
    * @return non-null list
    */
    public static Set makeSet(Object[] items)
    {
        Set ret = new HashSet(Math.max(items.length,1));
        for(int i = 0; i < items.length; i++)
            ret.add(items[i]);
        return(ret);
    }
    /**
    * Make a differnce set - only things in items not in remove
    * @param items non-null array to remove from
    * @param remove non-null array to remove from items
    * @return non-null array of the same type as items
    */
    public static Set makeDifferenceSet(Object[] items,Object[] remove)
    {
        Set holder = makeSet(items);
        for(int i = 0; i < remove.length; i++)
            holder.remove(remove[i]);
        return(holder);
    }
    /**
    * Make a differnce Array - only things in items not in remove
    * @param items non-null array to remove from
    * @param remove non-null array to remove from items
    * @return non-null array of the same type as items
    */
    public static Object[] makeDifferenceArray(Object[] items,Object[] remove)
    {
        if(items == null)
            return(EMPTY_OBJECT_ARRAY);
        if(remove == null)
            return(items);
        Set holder = makeDifferenceSet(items,remove);
        Object[] ret = (Object[])Array.newInstance(items.getClass().getComponentType(),holder.size());
        holder.toArray(ret);
        return(ret);
    }
    
    /**
    * Make an array from the contents of an enumeration
    * @param enum non-null enumeration
    * @return non-null List holding the elements of enum
    */
   public static List enumToList(Enumeration enu)
    {
        List holder = new ArrayList(256);
        while(enu.hasMoreElements())
            holder.add(enu.nextElement());
        return(holder);
    }
    /**
    * Make an array from the contents of an enumeration
    * @param enum non-null enumeration
    * @return non-null array holding the elements of enum
    */
   public static Object[] enumToArray(Enumeration enu)
    {
        List holder = enumToList(enu);
        Object[] ret = new Object[holder.size()];
        holder.toArray(ret);
        return(ret);
    }
    /**
    * Make an array from the contents of an enumeration
    * @param enum non-null Iterator
    * @return non-null List holding the elements of enum
    */
   public static List iteratorToList(Iterator it)
    {
        List holder = new ArrayList(256);
        while(it.hasNext())
            holder.add(it.next());
        return(holder);
    }
    /**
    * Make an array from the contents of an enumeration
    * @param enum non-null Iterator
    * @return non-null array holding the elements of enum
    */
   public static Object[] iteratorToArray(Iterator it)
    {
        List holder = iteratorToList(it);
        Object[] ret = new Object[holder.size()];
        holder.toArray(ret);
        return(ret);
    }
    /**
    * find all strings in a string array with a specific enf string
    * @param data non-null array of string
    * @param test possibly null test string null returns all
    * @return non-null array holding the matching strings
    */
    public static String[] getStringsEndingWith(String[] data,String test)
    {
        if(test == null || test.length() == 0)
            return(data);
        List holder = new ArrayList(data.length);
        for(int i = 0; i < data.length; i++) {
            if(data[i].endsWith(test))
                holder.add(data[i]);
        }
        return((String[])Util.collectionToArray(holder, String.class));
    }
    
    /**
    * find all strings in a string array with a specific start string
    * @param data non-null array of string
    * @param test possibly null test string null returns all
    * @return non-null array holding the matching strings
    */
    public static String[] getStringsStartingWith(String[] data,String test)
    {
        if(test == null || test.length() == 0)
            return(data);
        List holder = new ArrayList(data.length);
        for(int i = 0; i < data.length; i++) {
            if(data[i].startsWith(test))
                holder.add(data[i]);
        }
        return((String[])Util.collectionToArray(holder, String.class));
    }
    
    /**
    * convets value into en even percentage of 
    * total i.e. 0,10,20 ...
    * @param value positive int <= total
    * @param total positive int
    * @return valus as an even percentage of total
    */
    public static int decile(int value,int total)
    {
        if(value <= 0 || total <= 0)
            return(0);
        int test = (100 * value) / total; // percent
        int decile = (test + 4) / 10;
        int ret = Math.max(0,Math.min(100,decile * 10));
        return(ret);
    }
    
    /**
    * convert a Calendar into a readable string
    * @param in non-null Calandar
    * @return non-null String
    */
    public static String formatDate(java.util.Calendar in)
    {
        return(formatDate(in.getTime()));
    }
    
    /**
    * convert a Date into a readable string
    * @param in non-null Date
    * @return non-null String
    */
    public static String formatDate(java.util.Date in)
    {
        SimpleDateFormat fmt = new SimpleDateFormat("dd-MMM-yyyy HH:mm");
        return(fmt.format(in));
    }
    
    
    /**
    * write a string representing Now
    * @return non-null String
    */
    public static String nowString()
    {
        SimpleDateFormat fmt = new SimpleDateFormat("dd-MMM-yyyy HH:mm");
        return(fmt.format(new Date()));
    }
    
    
    public static class StringComparator implements Comparator
    {
        public int compare(Object o1,Object o2) {
            return(((String)o1).compareTo((String)o2));
        }
        public boolean equals(Object o1,Object o2) {
            return(((String)o1).equals(o2));
        }
    }
    
    public static class DoubleComparator implements Comparator
    {
        public int compare(Object o1,Object o2) {
            return(((Double)o1).compareTo((Double)o2));
        }
        public boolean equals(Object o1,Object o2) {
            return(((Double)o1).equals(o2));
        }
    }

    public static class IntegerComparator implements Comparator
    {
        public int compare(Object o1,Object o2) {
            return(((Integer)o1).compareTo((Integer)o2));
        }
        public boolean equals(Object o1,Object o2) {
            return(((Integer)o1).equals(o2));
        }
    }
    public static class DateComparator implements Comparator
    {
        public int compare(Object o1,Object o2) {
            return(((Date)o1).compareTo((Date)o2));
        }
        public boolean equals(Object o1,Object o2) {
            return(((Date)o1).equals(o2));
        }
    }
    
    public static void dumpMap(PrintStream out, Map Values) 
    {
        SortedSet keys = new TreeSet(Values.keySet());
        Iterator it = keys.iterator();
        while(it.hasNext()) {
            Object key = it.next();
            String keyStr = key.toString();
            String value = Values.get(key).toString();
            out.println("key = " + keyStr + " value = " + value);
        }
    }
  
  public static synchronized void setIPAddress(String in)
  {
        gLocalIPAddress  = in;
  }
    
  public static synchronized String getIPAddress()
  {
    if(gLocalIPAddress != null)
        return(gLocalIPAddress);
    try {
      InetAddress address = InetAddress.getLocalHost();
      return(address.getHostAddress());
    }
    catch (UnknownHostException e) {
      return("");
    }
  }
  
  /**
  * sort an array using the comparator 
  * return a new sorted array of the same type
  * @param non-null array 
  * @param c non-null comparator on the elements of in
  * @return non-null array
  */
  public static Object[] sortArray(Object[] in,Comparator c)
  {
    if(in.length < 2)
        return(in);
     SortedSet holder = new TreeSet(c);
     for(int i = 0; i < in.length; i++)
        holder.add(in[i]);
     Object[] out = (Object[])Array.newInstance(in.getClass().getComponentType(),holder.size());
     holder.toArray(out);
     return(out);
  }
  
  /*
  * return the class executing the code
  * @return non-null class
  */
  public Class executingClass() {
    return(gRevealer.doGetExecutingClass());
  }
  
  /*
  * true if this method is called from the named clazss
  */
  public boolean calledFromClass(Class Test) {
     Class[] TheStack = getStackClasses(); 
     for(int i = 0; i < TheStack.length; i++) {
        if(Test == TheStack[i])
            return(true);
     }
     return(false);
  }
  
    /**
    * hack to expose SecurityManager getClassContext method
    * to find what classes are executing 
    * can tell what class static methos is called
    * return non-null stack  class array
    */
    public Class[] getStackClasses() 
    {
        return(gRevealer.doGetClassContext());
    }
  
   /**
   * print a stack trace
   */
    public static String getStackTrace() {
        return(getStackTrace(new Throwable()));
    }
   /**
   * print a stack trace
   */
    public static String getStackTrace(Throwable err) {
        StringWriter strWrite = new StringWriter();
        PrintWriter prtWrite = new PrintWriter(strWrite);
        err.printStackTrace(prtWrite);
        StringBuffer temp = strWrite.getBuffer();
        String str = temp.toString();
        return str;
    }

//*******************************
//- *******************
//- End Class Util
    protected static StackRevealer gRevealer = new StackRevealer();
    
    /**
    * hack to expose SecurityManager getClassContext method
    * to find what classes are executing 
    * can tell what class static methos is called
    */
    protected static class StackRevealer extends SecurityManager
    {
        public  Class doGetExecutingClass() 
        {
            Class[] allClasses = getClassContext();
            return(allClasses[1]);
        }
        
        public  Class[] doGetClassContext() {
            Class[] allClasses = getClassContext();
            Class[] ret = new Class[allClasses.length - 1];
            System.arraycopy(allClasses, 1, ret, 0, allClasses.length - 1);
            return(ret);
        }
        
    }
    
}
